"""Simulator: Urban Simulator"""

import asyncio
import os
from datetime import datetime
from subprocess import Popen
from typing import Any, List, Literal, Optional, Tuple, Union

import aiohttp
import shapely
from agentsociety2.contrib.env.mobility_space.download_sim import download_binary
from agentsociety2.contrib.env.mobility_space.map import Map
from agentsociety2.contrib.env.mobility_space.utils import (
    find_free_ports,
    wait_for_port,
)
from agentsociety2.env import (
    EnvBase,
    tool,
)
from agentsociety2.logger import get_logger

from .replay_tables import AGENT_POSITION_SCHEMA
from pycityproto.city.geo.v2 import geo_pb2 as geo_pb2
from pycityproto.city.map.v2 import map_pb2 as map_pb2
from pycityproto.city.trip.v2.trip_pb2 import TripMode
from pydantic import BaseModel, ConfigDict, Field
from shapely import wkt
from shapely.geometry import LineString

__all__ = [
    "MobilitySpace",
    "MobilityPerson",
    "MobilityPersonInit",
    "Position",
    "Target",
    "Poi",
]


POI_START_ID = 7_0000_0000
DRIVING_SPEED_RATIO = 0.8  # the speed of driving is 80% of the max speed of the road
WALKING_SPEED = 1.34  # the speed of walking is 1.34 m/s
DRIVING_SPEED = 8.0  # the speed of driving is 8.0 m/s (approx 28.8 km/h)


class PositionInit(BaseModel):
    aoi_id: int = Field(
        ..., description="AOI ID, which is a continuous integer starting from 500000000"
    )
    poi_id: Optional[int] = Field(
        None,
        description="POI ID, which is a continuous integer starting from 700000000 (optional when initializing a person)",
    )


class Position(BaseModel):
    kind: Literal["aoi", "lane"] = Field(
        ..., description="Position kind: 'aoi' or 'lane'"
    )
    aoi_id: Optional[int] = Field(
        None,
        description="AOI ID, which is a continuous integer starting from 500000000",
    )
    poi_id: Optional[int] = Field(
        None,
        description="POI ID, which is a continuous integer starting from 700000000 (optional when initializing a person)",
    )
    xy: Tuple[float, float] = Field(..., description="XY coordinates of the position")
    lnglat: Tuple[float, float] = Field(
        ..., description="Lnglat coordinates of the position"
    )


class MobilityPersonInit(BaseModel):
    """Simplified model for initializing MobilityPerson, containing only id and position."""

    id: int = Field(..., description="Person ID")
    position: PositionInit = Field(..., description="The position of the person.")


class Target(BaseModel):
    position: Position = Field(..., description="The target position of the person.")
    mode: Literal["walking", "driving"] = Field(
        ..., description="The mode of the person."
    )
    path: shapely.LineString = Field(..., description="The path of the person.")
    path_s: float = Field(
        ..., description="The s coordinate of the person on the path."
    )
    path_v: float = Field(..., description="The speed of the person on the path.")

    model_config = ConfigDict(arbitrary_types_allowed=True)


class MobilityPerson(BaseModel):
    id: int = Field(..., description="Person ID")
    status: Literal["idle", "moving"] = Field(
        "idle",
        description='Person status. Default: "idle". If person is moving, it should be "moving".',
    )
    position: Position = Field(..., description="The position of the person.")
    target: Optional[Target] = Field(None, description="The target of the person.")

    model_config = ConfigDict(arbitrary_types_allowed=True)


# Response models for tool functions
class TargetResponse(BaseModel):
    """Response model for target information (without internal path details)"""

    position: Position = Field(..., description="The target position of the person")
    mode: Literal["walking", "driving"] = Field(
        ..., description="The mode of the person"
    )


class GetPersonResponse(BaseModel):
    """Response model for get_person() function"""

    id: int = Field(..., description="Person ID")
    status: Literal["idle", "moving"] = Field(..., description="Person status")
    position: Position = Field(..., description="The position of the person")
    target: Optional[TargetResponse] = Field(
        None, description="The target of the person"
    )


class Poi(BaseModel):
    """Point of Interest (POI) model"""

    id: int = Field(..., description="POI ID")
    name: str = Field(..., description="POI name")
    category: str = Field(..., description="POI category")
    position: dict = Field(..., description="POI position (x, y coordinates)")
    distance: Optional[float] = Field(
        None, description="Distance from search center (only for find_nearby_pois)"
    )


class FindNearbyPoisResponse(BaseModel):
    """Response model for find_nearby_pois() function"""

    pois: List[Poi] = Field(..., description="List of POIs found")


class MobilitySpace(EnvBase):
    """
    The environment, including map data, simulator clients, and environment variables.
    """

    TRIPMODE2STR = {
        TripMode.TRIP_MODE_WALK_ONLY: "walking",
        TripMode.TRIP_MODE_DRIVE_ONLY: "driving",
    }

    def __init__(
        self,
        file_path: str,
        home_dir: str,
        persons: List[MobilityPersonInit] | List[dict],
        poi_search_limit: int = 10,
    ):
        """
        Initialize the Environment.

        Args:
            file_path: The path to the map file.
            home_dir: The home directory of the environment.
            persons: The persons to initialize the environment with. Can be a list of MobilityPersonInit objects or dicts.
            poi_search_limit: The limit of POIs to search.
        """
        super().__init__()

        # Expand ~ in paths
        file_path = os.path.expanduser(file_path)
        home_dir = os.path.expanduser(home_dir)

        os.makedirs(home_dir, exist_ok=True)
        self._routing_bin_path = download_binary(home_dir)
        self._file_path = file_path
        self._home_dir = home_dir
        self._map = Map(file_path)
        # type annotation
        self._routing_proc: Optional[Popen] = None

        self.poi_id_2_aoi_id: dict[int, int] = {
            poi["id"]: poi["aoi_id"] for poi in self._map._poi_list
        }
        self._poi_search_limit = poi_search_limit
        """limit of POIs to search"""

        self._lock = asyncio.Lock()
        """lock for routing process"""

        # 位置跟踪存储（用于benchmark数据收集）
        self._person_trajectories: dict[int, list[Tuple[float, float]]] = {
            p["id"] if isinstance(p, dict) else p.id: [] for p in persons
        }
        """Store trajectory points for each person"""

        self._person_visited_aois: dict[int, set[int]] = {
            p["id"] if isinstance(p, dict) else p.id: set() for p in persons
        }
        """Store visited AOIs for each person"""

        # data: convert MobilityPersonInit or dict to MobilityPerson
        person_objects = []
        for p in persons:
            if isinstance(p, dict):
                # Convert MobilityPersonInit dict to MobilityPerson
                person_init = MobilityPersonInit.model_validate(p)
            else:
                person_init = p
            # convert PositionInit to Position
            if person_init.position.poi_id is not None:
                poi = self._map.get_poi(person_init.position.poi_id)
                assert poi is not None
                x, y = poi["position"]["x"], poi["position"]["y"]
            else:
                aoi = self._map.get_aoi(person_init.position.aoi_id)
                assert aoi is not None
                x, y = aoi["shapely_xy"].centroid.x, aoi["shapely_xy"].centroid.y
            position = Position(
                kind="aoi",
                aoi_id=person_init.position.aoi_id,
                poi_id=person_init.position.poi_id,
                xy=(x, y),
                lnglat=self._map.projector(x, y, inverse=True),
            )
            person_objects.append(
                MobilityPerson(
                    id=person_init.id,
                    status="idle",
                    position=position,
                    target=None,
                )
            )

        self._persons: dict[int, MobilityPerson] = {p.id: p for p in person_objects}
        
        # Step counter for replay data
        self._step_counter: int = 0

    def get_aoi_as_position(self, aoi_id: int) -> Position:
        """
        Get the position of an AOI.
        """
        aoi = self._map.get_aoi(aoi_id)
        assert aoi is not None
        x, y = aoi["shapely_xy"].centroid.x, aoi["shapely_xy"].centroid.y
        return Position(
            kind="aoi",
            aoi_id=aoi_id,
            poi_id=None,
            xy=(x, y),
            lnglat=self._map.projector(x, y, inverse=True),
        )

    def get_poi_as_position(self, poi_id: int) -> Position:
        """
        Get the position of a POI.
        """
        poi = self._map.get_poi(poi_id)
        assert poi is not None
        x, y = poi["position"]["x"], poi["position"]["y"]
        return Position(
            kind="aoi",
            aoi_id=poi["aoi_id"],
            poi_id=poi_id,
            xy=(x, y),
            lnglat=self._map.projector(x, y, inverse=True),
        )

    # ============================================================================
    # Probe Functions for Trajectory and Position Tracking (for Benchmark)
    # ============================================================================

    def record_person_position(self, person_id: int):
        """
        Record the current position of a person to their trajectory.

        Args:
            person_id: The ID of the person
        """
        if person_id not in self._persons:
            return

        # Initialize if not already present（处理动态添加的person）
        if person_id not in self._person_trajectories:
            self._person_trajectories[person_id] = []
        if person_id not in self._person_visited_aois:
            self._person_visited_aois[person_id] = set()

        person = self._persons[person_id]

        # Record XY coordinates
        xy = person.position.xy
        if xy not in self._person_trajectories[person_id]:
            self._person_trajectories[person_id].append(xy)

        # Record AOI if at an AOI
        if person.position.aoi_id is not None:
            self._person_visited_aois[person_id].add(person.position.aoi_id)

    def get_person_trajectory(self, person_id: int) -> list[Tuple[float, float]]:
        """
        Get the recorded trajectory of a person.

        Args:
            person_id: The ID of the person

        Returns:
            List of (x, y) coordinates
        """
        return self._person_trajectories.get(person_id, [])

    def get_person_visited_aois(self, person_id: int) -> set[int]:
        """
        Get the set of AOIs visited by a person.

        Args:
            person_id: The ID of the person

        Returns:
            Set of AOI IDs
        """
        return self._person_visited_aois.get(person_id, set())

    def get_all_persons_trajectories(self) -> dict[int, list[Tuple[float, float]]]:
        """
        Get trajectories for all persons.

        Returns:
            Dict mapping person_id to list of (x, y) coordinates
        """
        return self._person_trajectories.copy()

    def get_all_persons_visited_aois(self) -> dict[int, set[int]]:
        """
        Get visited AOIs for all persons.

        Returns:
            Dict mapping person_id to set of AOI IDs
        """
        return {pid: aois.copy() for pid, aois in self._person_visited_aois.items()}

    @classmethod
    def mcp_description(cls) -> str:
        """
        Return a description text for MCP environment module candidate list.
        Includes parameter descriptions and JSON schemas for data models.
        """
        import json

        # Get JSON schemas for nested models
        person_init_schema = MobilityPersonInit.model_json_schema()

        description = f"""{cls.__name__}: Mobility management module for urban navigation and location tracking.

**Description:** {cls.__doc__ or 'No description available'}

**Initialization Parameters (excluding llm):**
- file_path (str): The path to the map file.
- home_dir (str): The home directory of the environment.
- persons (List[MobilityPersonInit] | List[dict]): List of persons to initialize the environment with. Can be MobilityPersonInit objects or dicts matching the schema.
- poi_search_limit (int, optional): The limit of POIs to search. Default: 10.

**MobilityPersonInit JSON Schema:**
```json
{json.dumps(person_init_schema, indent=2)}
```

**Example initialization config:**
```json
{{
  "file_path": "/path/to/map.pb",
  "home_dir": "/path/to/home",
  "persons": [
    {{
      "id": 1,
      "position": {{
        "aoi_id": 500000000
      }}
    }}
  ],
  "poi_search_limit": 10
}}
```
"""
        return description

    @property
    def description(self):
        """Description of the environment module for router selection and function calling"""
        return """Mobility management module for urban navigation and location tracking.

**Available Operations:**
1. Person Location & Status: Query current position, movement state, and trip progress by calling the `get_person` function
2. Move to: Plan and execute trips to POIs/AOIs with walking or driving modes by calling the `move_to` function.
3. Find Nearby POIs: Search nearby places by category within specified radius by calling the `find_nearby_pois` function
4. Get POI Details: Get information about specific POIs by calling the `get_poi` function

**Key Concepts:**
- AOI (Area of Interest): Large city zones like districts or neighborhoods (identified by aoi_id)
- POI (Point of Interest): Specific locations like shops, restaurants, landmarks (identified by poi_id, starts from 700000000)
- Categories: Two-level hierarchy (first-level: 'restaurant', 'outdoor_activity'; second-level: 'cafe', 'park')
- Travel Modes: 'walking' or 'driving'
- Coordinates: Longitude/latitude pairs for precise positioning

**Usage Tips:**
- Use category filters to narrow POI searches
"""

    async def init(self, start_datetime: datetime) -> Any:
        """
        Initialize the environment including the routing.
        """
        await super().init(start_datetime)
        # =========================
        # init syncer
        # =========================
        routing_port = find_free_ports()[0]
        self._server_addr = f"localhost:{routing_port}"
        self._routing_proc = Popen(
            [
                self._routing_bin_path,
                "-listen",
                self._server_addr,
                "-map",
                self._file_path,
                "-log-level",
                "warn",
            ],
            env=os.environ,
        )

        get_logger().info(
            f"start routing at {self._server_addr}, PID={self._routing_proc.pid}"
        )

        # Wait for the routing server to start listening
        get_logger().info(
            f"Waiting for routing server to start on port {routing_port}..."
        )
        if not wait_for_port("localhost", routing_port, timeout=30.0):
            # If the port is not available, kill the process and raise an error
            if self._routing_proc.poll() is None:
                self._routing_proc.kill()
            raise RuntimeError(
                f"Routing server failed to start on port {routing_port} within 30 seconds"
            )
        get_logger().info(f"Routing server is ready on {self._server_addr}")

        if self._replay_writer is not None:
            await self._replay_writer.register_table(AGENT_POSITION_SCHEMA)
            get_logger().info("Registered agent_position table for MobilitySpace")

    async def get_agent_position(self, agent_id: int) -> Tuple[Optional[float], Optional[float]]:
        """Get the position of an agent.

        Args:
            agent_id: The ID of the agent (person).

        Returns:
            A tuple of (longitude, latitude), or (None, None) if not available.
        """
        if agent_id not in self._persons:
            return None, None
        person = self._persons[agent_id]
        lng, lat = person.position.lnglat
        return lng, lat

    async def _write_position(
        self, agent_id: int, step: int, t: datetime, lng: float, lat: float
    ) -> None:
        """Write agent position to the agent_position table.

        Args:
            agent_id: The agent ID.
            step: The simulation step number.
            t: The simulation datetime.
            lng: Longitude coordinate.
            lat: Latitude coordinate.
        """
        if self._replay_writer is not None:
            await self._replay_writer.write("agent_position", {
                "id": agent_id,
                "step": step,
                "t": t,
                "lng": lng,
                "lat": lat,
            })


    @property
    def map(self):
        assert self._map is not None, "Map not initialized"
        return self._map

    @property
    def projector(self):
        return self._map.projector

    def _get_around_pois(
        self,
        center: tuple[float, float],
        radius: Optional[float] = None,
        poi_type: Optional[Union[str, list[str]]] = None,
        limit: int = 10,
    ) -> list[dict]:
        """
        Get Points of Interest (POIs) around a central point based on type.

        - **Args**:
            - `center` (`Tuple[float, float]`): The central point as a tuple.
            - `radius` (`Optional[float]`): The search radius in meters. If not provided, all POIs are considered.
            - `poi_type` (`Optional[Union[str, List[str]]]`): The category or categories of POIs to filter by. Can be first-level categories, second-level categories, or None for all POIs.

        - **Returns**:
            - `List[Dict]`: A list of dictionaries containing information about the POIs found.
        """
        # If no poi_type specified, return all POIs
        if poi_type is None:
            assert self._map is not None
            all_pois: list[tuple[dict, float]] = self._map.query_pois(
                center=center,
                radius=radius,
            )
            pois = [{"poi": poi, "distance": distance} for poi, distance in all_pois]
            return pois[:limit]

        # Process poi_type input
        if isinstance(poi_type, str):
            poi_type = [poi_type]

        transformed_poi_type: list[str] = []
        for t in poi_type:
            if t in self._map.poi_cate:
                # This is a first-level category, expand to all subcategories
                transformed_poi_type += self._map.poi_cate[t]
            else:
                # This is either a second-level category or unknown category
                transformed_poi_type.append(t)

        poi_type_set = set(transformed_poi_type)

        # query pois within the radius
        assert self._map is not None
        nearest_pois: list[tuple[dict, float]] = self._map.query_pois(
            center=center,
            radius=radius,
        )

        # Filter POIs by category
        pois = []
        for poi, distance in nearest_pois:
            catg = poi["category"]
            # Check if any category in the POI matches our filter
            if any(c in poi_type_set for c in catg):
                pois.append({"poi": poi, "distance": distance})

        return pois[:limit]

    # ============================================================================
    # Mobility Management Functions for LLM Function Calling
    # ============================================================================

    @tool(readonly=True, kind="observe")
    async def get_person(self, person_id: int) -> GetPersonResponse:
        """
        Get the current location and status of a person, including position, movement state, and trip progress.

        Args:
            person_id: The ID of the person to query

        Returns:
            The context containing detailed location and movement information
        """
        if person_id not in self._persons:
            raise ValueError(f"Person {person_id} not found")

        person = self._persons[person_id]

        # Construct target response if target exists
        target_response = None
        if person.target is not None:
            target_response = TargetResponse(
                position=person.target.position,
                mode=person.target.mode,
            )

        return GetPersonResponse(
            id=person.id,
            status=person.status,
            position=person.position,
            target=target_response,
        )

    @tool(readonly=False)
    async def move_to(
        self,
        person_id: int,
        aoi_id_or_poi_id: int,
        mode: Literal["walking", "driving"] = "driving",
    ):
        """
        Plan and start a trip for a person to move to a specific AOI or POI.

        Args:
            person_id: The ID of the person to move
            aoi_id_or_poi_id: The target location ID (AOI ID or POI ID starting from 700000000)
            mode: The travel mode (walking or driving)

        Returns:
            Empty response indicating success
        """
        if person_id not in self._persons:
            raise ValueError(f"Person {person_id} not found")

        person = self._persons[person_id]

        # 1. choose mode

        # choose the start position
        if person.position.kind == "aoi":
            start_position = {
                "aoi_position": {
                    "aoi_id": person.position.aoi_id,
                }
            }
        else:
            # search for nearest road
            radius = 1000
            lanes = []
            while radius < 10000:
                lanes = self._map.query_lane(
                    person.position.xy, radius, 1 if mode == "driving" else 2
                )
                if len(lanes) > 0:
                    break
                radius *= 2
            if len(lanes) == 0:
                return
            lane, s, _ = lanes[0]
            start_position = {
                "lane_position": {
                    "lane_id": lane["id"],
                    "s": s,
                }
            }

        # Get destination position
        if aoi_id_or_poi_id >= POI_START_ID:
            poi = self._map.get_poi(aoi_id_or_poi_id)
            if poi is None:
                return
            destination_position = {
                "aoi_position": {
                    "aoi_id": poi["aoi_id"],
                }
            }
        else:
            aoi = self._map.get_aoi(aoi_id_or_poi_id)
            if aoi is None:
                return
            destination_position = {
                "aoi_position": {
                    "aoi_id": aoi_id_or_poi_id,
                }
            }

        # Call routing service to get route
        url = f"http://{self._server_addr}/city.routing.v2.RoutingService/GetRoute"
        request_data = {
            "type": (
                TripMode.TRIP_MODE_DRIVE_ONLY
                if mode == "driving"
                else TripMode.TRIP_MODE_WALK_ONLY
            ),
            "start": start_position,
            "end": destination_position,
        }
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=request_data) as response:
                response_data = await response.json()
        eta = self._map.estimate_route_time(request_data, response_data)
        xys = self._map._route_to_xys(request_data, response_data)
        path = LineString(xys)
        path_v = path.length / eta if eta > 0 else 1

        # Update person state
        person.status = "moving"
        # Determine destination position with poi_id if applicable
        dest_aoi_id = destination_position["aoi_position"]["aoi_id"]
        dest_poi_id = aoi_id_or_poi_id if aoi_id_or_poi_id >= POI_START_ID else None
        if dest_poi_id is not None:
            # Use POI position
            poi = self._map.get_poi(dest_poi_id)
            assert poi is not None
            x, y = poi["position"]["x"], poi["position"]["y"]
            poi_xy = (x, y)
            poi_lnglat = self._map.projector(poi_xy[0], poi_xy[1], inverse=True)
            target_position = Position(
                kind="aoi",
                aoi_id=dest_aoi_id,
                poi_id=dest_poi_id,
                xy=poi_xy,
                lnglat=poi_lnglat,
            )
        else:
            # Use route end position (AOI centroid)
            target_position = Position(
                kind="aoi",
                aoi_id=dest_aoi_id,
                poi_id=None,
                xy=xys[-1],
                lnglat=self._map.projector(xys[-1][0], xys[-1][1], inverse=True),
            )

        person.target = Target(
            position=target_position,
            mode=mode,
            path=path,
            path_s=0.0,
            path_v=path_v,
        )
        xy = path.interpolate(0)
        person.position = Position(
            kind="lane",
            aoi_id=None,
            poi_id=None,
            xy=(xy.x, xy.y),
            lnglat=self._map.projector(xy.x, xy.y, inverse=True),
        )

        return

    @tool(readonly=True)
    async def find_nearby_pois(
        self, x: float, y: float, category: Optional[str], radius: float
    ) -> FindNearbyPoisResponse:
        """
        Discover Points of Interest (POIs) near a location, filtered by category and distance.

        Args:
            x: The longitude (x coordinate) of the search center
            y: The latitude (y coordinate) of the search center
            category: POI category filter - first-level (e.g., 'restaurant'), second-level (e.g., 'cafe'), or None for all categories.
                Available first-level categories (with examples):
                  - 'children_playground': playground, summer_camp, miniature_golf, dog_park
                  - 'cultural_and_artistic': arts_centre, brothel, casino, cinema, community_centre
                  - 'education_institution': college, dancing_school, driving_school, first_aid_school, kindergarten
                  - 'financial_service': atm, payment_terminal, bank, bureau_de_change, money_transfer
                  - 'indoor_entertainment': adult_gaming_centre, amusement_arcade, bowling_alley, disc_golf_course, escape_game
                  - 'medical_care': baby_hatch, clinic, dentist, doctors, hospital
                  - 'nature_and_wildlife_observation': bird_hide, nature_reserve, wildlife_hide, hunting_stand
                  - 'other_special_purpose': animal_boarding, animal_breeding, animal_shelter, animal_training, baking_oven
                  - 'outdoor_activity': bandstand, beach_resort, bird_hide, bleachers, firepit
                  - 'public_service': bbq, bench, dog_toilet, dressing_room, drinking_water
                  - 'restaurant': bar, biergarten, cafe, fast_food, food_court
                  - 'sports_facility': horse_riding, ice_rink, marina, pitch, sports_centre
                  - 'transportation_facility': bicycle_parking, bicycle_repair_station, bicycle_rental, bicycle_wash, boat_rental
                  - 'water_activity': beach_resort, ice_rink, marina, slipway, swimming_area
                You can also use any first-level category or second-level category directly (e.g., 'children_playground' 'cafe', 'park', 'hospital').
            radius: Search radius in meters (e.g., 1000 for 1km)

        Returns:
            Structured data containing POI list with IDs, names, categories, positions, and distances
        """
        pois = self._get_around_pois(
            center=(x, y),
            radius=radius,
            poi_type=category,
            limit=self._poi_search_limit,
        )
        clean_pois = []
        for p in pois:
            clean_poi = Poi(
                id=p["poi"]["id"],
                name=p["poi"]["name"],
                position=p["poi"]["position"],
                category=p["poi"]["category"][-1],
                distance=p["distance"],
            )
            clean_pois.append(clean_poi)

        return FindNearbyPoisResponse(pois=clean_pois)

    @tool(readonly=True)
    async def get_poi(self, poi_id: int) -> Poi:
        """
        Retrieve detailed information about a specific Point of Interest.

        Args:
            poi_id: The unique ID of the POI (starts from 700000000)

        Returns:
            Structured data containing POI details (ID, name, category, position)
        """
        poi = self._map.get_poi(poi_id)
        if poi is None:
            raise ValueError(f"POI {poi_id} not found")

        poi_obj = Poi(
            id=poi["id"],
            name=poi["name"],
            category=poi["category"][-1],
            position=poi["position"],
            distance=None,
        )

        return poi_obj

    async def close(self):
        """
        Terminate the simulation process if it's running.
        """
        if self._routing_proc is not None and self._routing_proc.poll() is None:
            get_logger().info(
                f"Terminating routing at {self._server_addr}, PID={self._routing_proc.pid}, please ignore the PANIC message"
            )
            self._routing_proc.kill()
            # wait for the process to terminate
            self._routing_proc.wait()
        self._routing_proc = None

    async def step(self, tick: int, t: datetime):
        """
        Run forward one step.

        Args:
            tick: The number of ticks (1 tick = 1 second) of this simulation step.
            t: The current datetime of the simulation after this step with the ticks.
        """
        # Update all moving persons
        for person in self._persons.values():
            if person.status == "moving":
                assert person.target is not None
                distance_to_move = person.target.path_v * tick
                person.target.path_s += distance_to_move
                if person.target.path_s >= person.target.path.length:
                    person.status = "idle"
                    person.position = person.target.position
                    person.target = None
                    # update position - use POI position if poi_id exists, otherwise use AOI centroid
                    assert person.position.aoi_id is not None
                    if person.position.poi_id is not None:
                        # Use POI position
                        poi = self._map.get_poi(person.position.poi_id)
                        assert poi is not None
                        x, y = poi["position"]["x"], poi["position"]["y"]
                        person.position.xy = (x, y)
                        person.position.lnglat = self._map.projector(x, y, inverse=True)
                    else:
                        # Use AOI centroid
                        aoi = self._map.get_aoi(person.position.aoi_id)
                        assert aoi is not None
                        x, y = (
                            aoi["shapely_xy"].centroid.x,
                            aoi["shapely_xy"].centroid.y,
                        )
                        person.position.xy = (x, y)
                        person.position.lnglat = self._map.projector(x, y, inverse=True)
                else:
                    # update position
                    xy = person.target.path.interpolate(person.target.path_s)
                    person.position.xy = (xy.x, xy.y)
                    person.position.lnglat = self._map.projector(
                        xy.x, xy.y, inverse=True
                    )

            # 【关键】每步记录person的位置（用于轨迹收集）
            self.record_person_position(person.id)
            
            # Write position to replay database
            lng, lat = person.position.lnglat
            await self._write_position(person.id, self._step_counter, t, lng, lat)

        self.t = t
        self._step_counter += 1

    # ==================== Replay Data Methods ====================
    # Note: get_agent_position and _write_position are defined earlier in the class

    def _dump_state(self) -> dict:
        """
        Dump the internal state of MobilitySpace for serialization.

        Returns:
            dict: A dictionary containing all necessary state information
        """
        # Serialize persons
        persons_data = {}
        for person_id, person in self._persons.items():
            person_dict = person.model_dump()
            # Convert shapely LineString to WKT (Well-Known Text) format if present
            if person.target is not None and person.target.path is not None:
                person_dict["target"]["path"] = person.target.path.wkt
            persons_data[person_id] = person_dict

        return {
            "file_path": self._file_path,
            "home_dir": self._home_dir,
            "poi_search_limit": self._poi_search_limit,
            "persons": persons_data,
        }

    def _load_state(self, state: dict):
        """
        Load the internal state of MobilitySpace from serialized data.

        Args:
            state: The state dictionary produced by _dump_state()
        """
        # Restore configuration parameters
        self._poi_search_limit = state.get("poi_search_limit", 10)

        # Restore persons
        persons_data = state.get("persons", {})
        self._persons = {}
        for person_id_str, person_dict in persons_data.items():
            person_id = int(person_id_str)
            # Convert WKT back to shapely LineString if present
            if (
                person_dict.get("target") is not None
                and person_dict["target"].get("path") is not None
            ):
                path_wkt = person_dict["target"]["path"]
                person_dict["target"]["path"] = wkt.loads(path_wkt)

            # Reconstruct MobilityPerson from dict
            person = MobilityPerson.model_validate(person_dict)
            self._persons[person_id] = person



"""Table schema definitions for dynamic replay table registration.

Environment modules (e.g. social_media, mobility_space) define their own replay tables
using TableSchema and ColumnDef. They call ReplayWriter.register_table(schema) to
create the table and ReplayWriter.write(table_name, row) to persist data. The replay
API reads these tables using query-only SQLAlchemy Table definitions provided by each
module in its replay_tables.py (e.g. contrib/env/social_media/replay_tables.py,
contrib/env/mobility_space/replay_tables.py). Framework tables (agent_profile,
agent_status, agent_dialog) are defined in models.py (SQLModel) and created by
ReplayWriter.init(); they are not registered via TableSchema.
"""

from dataclasses import dataclass, field
from typing import List, Literal, Optional, Tuple

# SQLite column types
ColumnType = Literal["INTEGER", "REAL", "TEXT", "BLOB", "TIMESTAMP", "JSON"]


@dataclass
class ColumnDef:
    """Column definition for a table.
    
    Attributes:
        name: Column name
        type: SQLite column type
        nullable: Whether the column allows NULL values (default True)
        default: Default value expression (e.g., "CURRENT_TIMESTAMP")
    """
    name: str
    type: ColumnType
    nullable: bool = True
    default: Optional[str] = None
    
    def to_sql(self) -> str:
        """Generate SQL column definition."""
        parts = [self.name, self.type]
        if not self.nullable:
            parts.append("NOT NULL")
        if self.default is not None:
            parts.append(f"DEFAULT {self.default}")
        return " ".join(parts)


@dataclass
class TableSchema:
    """Schema definition for a database table.
    
    This class allows environment modules to define their own tables
    that will be created dynamically by ReplayWriter.
    
    Attributes:
        name: Table name
        columns: List of column definitions
        primary_key: List of column names that form the primary key
        indexes: List of index definitions (column name or list of column names)
    
    Example:
        >>> schema = TableSchema(
        ...     name="agent_position",
        ...     columns=[
        ...         ColumnDef("id", "INTEGER", nullable=False),
        ...         ColumnDef("step", "INTEGER", nullable=False),
        ...         ColumnDef("t", "TIMESTAMP", nullable=False),
        ...         ColumnDef("lng", "REAL"),
        ...         ColumnDef("lat", "REAL"),
        ...     ],
        ...     primary_key=["id", "step"],
        ...     indexes=[["step"], ["t"]],
        ... )
    """
    name: str
    columns: List[ColumnDef]
    primary_key: List[str] = field(default_factory=list)
    indexes: List[List[str]] = field(default_factory=list)
    
    def to_create_sql(self) -> str:
        """Generate CREATE TABLE SQL statement."""
        column_defs = [col.to_sql() for col in self.columns]
        
        # Add primary key constraint
        if self.primary_key:
            pk_cols = ", ".join(self.primary_key)
            column_defs.append(f"PRIMARY KEY ({pk_cols})")
        
        columns_sql = ",\n    ".join(column_defs)
        return f"CREATE TABLE IF NOT EXISTS {self.name} (\n    {columns_sql}\n)"
    
    def to_index_sql(self) -> List[str]:
        """Generate CREATE INDEX SQL statements."""
        statements = []
        for idx_cols in self.indexes:
            idx_name = f"idx_{self.name}_{'_'.join(idx_cols)}"
            cols = ", ".join(idx_cols)
            statements.append(
                f"CREATE INDEX IF NOT EXISTS {idx_name} ON {self.name}({cols})"
            )
        return statements



"""Literature search and management module

Provides functionality for:
- Searching academic literature
- Saving and indexing literature in workspace
- Formatting literature entries
"""

from agentsociety2.skills.literature.models import LiteratureEntry, LiteratureIndex
from agentsociety2.skills.literature.formatter import (
    sanitize_filename,
    format_article_as_markdown,
)
from agentsociety2.skills.literature.search import (
    search_literature_and_save,
    generate_summary,
    format_search_results,
    load_literature_index,
)
from agentsociety2.skills.literature.core import (
    search_literature,
    filter_relevant_literature,
    format_literature_info,
    is_chinese_text,
)

__all__ = [
    # Models
    "LiteratureEntry",
    "LiteratureIndex",
    # Formatter
    "sanitize_filename",
    "format_article_as_markdown",
    # Search
    "search_literature_and_save",
    "generate_summary",
    "format_search_results",
    "load_literature_index",
    # Core
    "search_literature",
    "filter_relevant_literature",
    "format_literature_info",
    "is_chinese_text",
]



"""
FastAPI backend service for AI Social Scientist VSCode extension

关联文件：
- @packages/agentsociety2/agentsociety2/backend/run.py - 服务启动脚本
- @extension/src/services/backendManager.ts - VSCode插件后端进程管理
- @extension/src/apiClient.ts - VSCode插件API客户端

路由注册：
- @packages/agentsociety2/agentsociety2/backend/routers/prefill_params.py - /api/v1/prefill-params
- @packages/agentsociety2/agentsociety2/backend/routers/experiments.py - /api/v1/experiments
- @packages/agentsociety2/agentsociety2/backend/routers/replay.py - /api/v1/replay
- @packages/agentsociety2/agentsociety2/backend/routers/custom.py - /api/v1/custom
- @packages/agentsociety2/agentsociety2/backend/routers/modules.py - /api/v1/modules
"""

from __future__ import annotations

import os
import logging
from contextlib import asynccontextmanager
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from dotenv import load_dotenv
from pathlib import Path

from agentsociety2.backend.routers import prefill_params, experiments, replay, custom, modules

# 加载环境变量
_project_root = Path(__file__).resolve().parents[2]
load_dotenv(_project_root / ".env")


# 配置标准 logging
def _setup_logging():
    """配置应用日志"""
    log_level = os.getenv("BACKEND_LOG_LEVEL", "info")
    # 将 uvicorn 的 "trace" 映射到 Python logging 的 "DEBUG"
    python_log_level = "DEBUG" if log_level.lower() == "trace" else log_level.upper()
    level = getattr(logging, python_log_level, logging.INFO)

    # 配置根 logger（如果还没有配置过）
    root_logger = logging.getLogger()
    if not root_logger.handlers:
        logging.basicConfig(
            level=level,
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            force=True,  # Python 3.8+ 支持，强制重新配置
        )
    else:
        # 如果已经配置过，只更新日志等级
        root_logger.setLevel(level)

    # 设置 agentsociety2 相关模块的日志等级
    agentsociety_logger = logging.getLogger("agentsociety2")
    agentsociety_logger.setLevel(level)

    return agentsociety_logger


_setup_logging()
from agentsociety2.logger import get_logger

logger = get_logger()


@asynccontextmanager
async def lifespan(app: FastAPI):
    """应用生命周期管理"""
    # 启动时执行
    logger.info("AI Social Scientist Backend Service 启动中...")
    logger.info(f"项目根目录: {_project_root}")

    yield

    # 关闭时执行
    logger.info("AI Social Scientist Backend Service 关闭中...")


# 创建FastAPI应用
app = FastAPI(
    title="AI Social Scientist Backend API",
    description="Backend API service for AI Social Scientist VSCode extension",
    version="2.0.0",
    lifespan=lifespan,
)

# 配置CORS（允许VSCode插件跨域访问）
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # 生产环境应该限制为特定域名
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# 注册路由（仅保留必要的API）
app.include_router(prefill_params.router)
app.include_router(experiments.router, prefix="/api/v1")
app.include_router(replay.router, prefix="/api/v1")
app.include_router(custom.router)
app.include_router(modules.router)


@app.get("/")
async def root():
    """根路径，返回API信息"""
    return {
        "service": "AI Social Scientist Backend API",
        "version": "2.0.0",
        "status": "running",
        "endpoints": {
            "prefill_params": "/api/v1/prefill-params",
            "experiments": "/api/v1/experiments/{hypothesis_id}/{experiment_id}",
            "replay": "/api/v1/replay/{hypothesis_id}/{experiment_id}/*",
            "custom": "/api/v1/custom/*",
            "modules": "/api/v1/modules/*",
        },
    }


@app.get("/health")
async def health_check():
    """健康检查端点"""
    return {"status": "healthy"}


@app.exception_handler(Exception)
async def global_exception_handler(request: Request, exc: Exception):
    """全局异常处理器"""
    logger.error(f"未处理的异常: {exc}", exc_info=True)
    return JSONResponse(
        status_code=500,
        content={
            "success": False,
            "error": "Internal Server Error",
            "detail": str(exc),
        },
    )


if __name__ == "__main__":
    import uvicorn
    import argparse

    # 解析命令行参数
    parser = argparse.ArgumentParser(
        description="启动 AI Social Scientist Backend API 服务"
    )
    parser.add_argument(
        "--log-level",
        type=str,
        default=None,
        choices=["critical", "error", "warning", "info", "debug", "trace"],
        help="设置日志等级 (critical, error, warning, info, debug, trace)",
    )
    args = parser.parse_args()

    # 从环境变量读取配置，命令行参数优先
    host = os.getenv("BACKEND_HOST", "0.0.0.0")
    port = int(os.getenv("BACKEND_PORT", "8001"))
    log_level = args.log_level or os.getenv("BACKEND_LOG_LEVEL", "info")

    # 如果命令行参数设置了日志等级，更新环境变量并重新配置日志
    if args.log_level:
        os.environ["BACKEND_LOG_LEVEL"] = args.log_level
        _setup_logging()

    logger.info(f"启动服务器: http://{host}:{port}")
    logger.info(f"日志等级: {log_level}")
    uvicorn.run(
        "agentsociety2.backend.app:app",
        host=host,
        port=port,
        reload=False,  # 生产环境设为False
        log_level=log_level,
    )



"""
AgentSociety2 自定义模块包

此包用于存放用户自定义的 Agent 和环境模块代码。

目录结构:
- agents/    : 自定义 Agent 类
- envs/      : 自定义环境模块类

使用方法:
1. 在相应目录创建 .py 文件
2. 继承 AgentBase 或 EnvBase
3. 运行 VSCode 命令 "扫描自定义模块"
"""

__all__ = []



import random
import math
from datetime import datetime
from typing import List, Dict, Any, Optional, Set
from collections import defaultdict

from .models import Post, User


def _get_pretrained_algorithm(algorithm_name: str, model_path: Optional[str]):
    """
    根据算法名和模型路径加载预训练推荐模型。
    预训练方式：使用 recommendation 子模块训练后 save(path)，此处 load(path) 即可接入。

    Args:
        algorithm_name: 如 "mf", "ncf", "lightgcn" 等，需与 recommendation.algorithms 中实现一致。
        model_path: 预训练模型文件路径（如 .pkl 或 目录）。

    Returns:
        已 load 的 RecommenderAlgorithm 实例，若 model_path 为空则返回 None。
    """
    if not model_path:
        return None
    algo = None
    if algorithm_name == "mf":
        from .recommendation.algorithms.mf import MFRecommender, MFConfig
        algo = MFRecommender(config=MFConfig())
    # 可在此扩展: "ncf" -> NCFRecommender, "lightgcn" -> LightGCNRecommender 等
    if algo is None:
        return None
    try:
        algo.load(model_path)
    except Exception:
        return None
    return algo


class RecommendationEngine:
    """
    推荐算法引擎。支持规则算法（时间序、热度、Twitter 排序、随机）与预训练模型（如 MF）。
    预训练模型：在 recommendation 子模块中训练并 save(path)，构造时传入 model_path 与 algorithm 即可接入。
    """

    def __init__(
        self,
        model_path: Optional[str] = None,
        recommendation_algorithm: str = "mf",
    ):
        """
        Args:
            model_path: 预训练模型路径。非空时加载对应算法并可在 refresh_feed 中通过 algorithm="mf" 使用。
            recommendation_algorithm: 算法名，如 "mf"。需与 recommendation.algorithms 中实现一致。
        """
        self._model_algorithm = _get_pretrained_algorithm(recommendation_algorithm, model_path)
        self._model_algorithm_name = recommendation_algorithm if self._model_algorithm else None

    def get_model_algorithm_name(self) -> Optional[str]:
        """若已加载预训练模型，返回算法名（如 'mf'），否则返回 None。"""
        return self._model_algorithm_name

    def model_recommend(
        self,
        posts: List[Post],
        user_id: int,
        limit: int = 20,
        exclude_post_ids: Optional[Set[int]] = None,
    ) -> List[Post]:
        """
        使用预训练模型对候选帖子排序。未加载模型时回退为时间序。

        Args:
            posts: 候选帖子列表（item_id = post_id）
            user_id: 用户 ID
            limit: 返回条数
            exclude_post_ids: 需要排除的 post_id 集合（如已读）

        Returns:
            按模型打分排序的帖子列表
        """
        if not posts:
            return []
        exclude = set(exclude_post_ids or [])
        post_by_id = {p.post_id: p for p in posts}
        candidate_ids = [p.post_id for p in posts]

        if self._model_algorithm is None:
            return self.chronological(posts, user_id, limit)

        try:
            rec_list = self._model_algorithm.recommend(user_id, limit, exclude)
        except Exception:
            return self.chronological(posts, user_id, limit)

        out: List[Post] = []
        added_ids: Set[int] = set()
        for item_id, _score in rec_list:
            if item_id in post_by_id and item_id not in exclude and item_id not in added_ids:
                out.append(post_by_id[item_id])
                added_ids.add(item_id)
            if len(out) >= limit:
                break
        # 若模型返回不足或含大量未在候选中的 id，用候选集顺序补足
        for pid in candidate_ids:
            if len(out) >= limit:
                break
            if pid not in added_ids:
                out.append(post_by_id[pid])
                added_ids.add(pid)
        return out[:limit]

    @staticmethod
    def chronological(
        posts: List[Post],
        user_id: int,
        limit: int = 20
    ) -> List[Post]:
        """
        时间序列排序算法（基线）
        
        Args:
            posts: 所有候选帖子列表
            user_id: 当前用户ID
            limit: 返回帖子数量
            
        Returns:
            按时间倒序排列的帖子列表
        """
        # 按创建时间降序排序
        sorted_posts = sorted(posts, key=lambda p: p.created_at, reverse=True)
        return sorted_posts[:limit]
    
    @staticmethod
    def reddit_hot(
        posts: List[Post],
        user_id: int,
        limit: int = 20
    ) -> List[Post]:
        """
        Reddit热度算法
        
        基于点赞数和时间衰减计算热度分数。
        公式: score = sign * log10(|likes|) + (created_at_seconds - epoch) / 45000
        
        参考: https://medium.com/hacking-and-gonzo/how-reddit-ranking-algorithms-work-ef111e33d0d9
        
        Args:
            posts: 所有候选帖子列表
            user_id: 当前用户ID
            limit: 返回帖子数量
            
        Returns:
            按热度分数排序的帖子列表
        """
        def calculate_hot_score(post: Post) -> float:
            """
            计算帖子的Reddit热度分数
            
            注意：这里简化了公式，只考虑点赞数
            """
            # 点赞数（没有dislike，所以s = num_likes）
            s = post.likes_count
            
            # log10(|s|) with minimum of 1 to avoid log(0)
            order = math.log10(max(abs(s), 1))
            
            sign = 1 if s > 0 else 0 if s == 0 else -1
            
            # 时间部分
            epoch = datetime(1970, 1, 1)
            td = post.created_at - epoch
            epoch_seconds = td.days * 86400 + td.seconds + (td.microseconds / 1e6)
            
            # Reddit epoch offset (2005-12-08)
            reddit_epoch = 1134028003
            seconds = epoch_seconds - reddit_epoch
            
            # 最终分数
            score = sign * order + seconds / 45000
            
            return round(score, 7)
        
        # 计算所有帖子的热度分数
        post_scores = [(post, calculate_hot_score(post)) for post in posts]
        
        # 按热度分数降序排序
        post_scores.sort(key=lambda x: x[1], reverse=True)
        
        # 返回top N
        return [post for post, score in post_scores[:limit]]
    
    @staticmethod
    def twitter_ranking(
        posts: List[Post],
        user_id: int,
        limit: int = 20,
        follows: Dict[int, List[int]] = None,
        likes: Dict[int, List[int]] = None,
        weights: Dict[str, float] = None
    ) -> List[Post]:
        """
        Twitter排序算法
        
        综合多个因素的加权排序：
        - 是否关注作者（following_weight）
        - 互动数（likes + reposts + comments）（engagement_weight）
        - 新鲜度（时间）（recency_weight）
        - 互动率（engagement / views）（engagement_rate_weight）
        
        Args:
            posts: 所有候选帖子列表
            user_id: 当前用户ID
            limit: 返回帖子数量
            follows: 关注关系 {follower_id: [followee_ids]}
            likes: 点赞记录 {post_id: [user_ids]}
            weights: 权重配置
            
        Returns:
            按综合分数排序的帖子列表
        """
        # 默认权重
        default_weights = {
            "following": 0.4,      # 关注权重
            "engagement": 0.3,     # 互动数权重
            "recency": 0.2,        # 新鲜度权重
            "engagement_rate": 0.1 # 互动率权重
        }
        
        if weights:
            default_weights.update(weights)
        
        follows = follows or {}
        likes = likes or {}
        
        # 用户关注的人
        following_ids = follows.get(user_id, [])
        
        # 当前时间
        max_time = max((p.created_at for p in posts), default=datetime.now())
        
        def calculate_twitter_score(post: Post) -> float:
            """计算Twitter排序分数"""
            score = 0.0
            
            # 1. 关注因素
            if post.author_id in following_ids:
                score += default_weights["following"]
            
            # 2. 互动数因素（归一化到0-1）
            total_engagement = post.likes_count + post.reposts_count + post.comments_count
            # 假设最大互动数为100
            engagement_normalized = min(total_engagement / 100.0, 1.0)
            score += default_weights["engagement"] * engagement_normalized
            
            # 3. 新鲜度因素（时间差越小，分数越高）
            time_diff_hours = (max_time - post.created_at).total_seconds() / 3600
            # 假设24小时内的帖子有新鲜度加成
            recency_score = max(0, 1 - time_diff_hours / 24.0)
            score += default_weights["recency"] * recency_score
            
            # 4. 互动率因素
            if post.view_count > 0:
                engagement_rate = total_engagement / post.view_count
                engagement_rate_normalized = min(engagement_rate, 1.0)
                score += default_weights["engagement_rate"] * engagement_rate_normalized
            
            return score
        
        # 计算所有帖子的分数
        post_scores = [(post, calculate_twitter_score(post)) for post in posts]
        
        # 按分数降序排序
        post_scores.sort(key=lambda x: x[1], reverse=True)
        
        # 返回top N
        return [post for post, score in post_scores[:limit]]
    
    @staticmethod
    def random_recommend(
        posts: List[Post],
        user_id: int,
        limit: int = 20
    ) -> List[Post]:
        """
        随机推荐算法（基线/对照组）
        
        Args:
            posts: 所有候选帖子列表
            user_id: 当前用户ID
            limit: 返回帖子数量
            
        Returns:
            随机选择的帖子列表
        """
        if len(posts) <= limit:
            return list(posts)
        else:
            return random.sample(posts, limit)


__all__ = ["RecommendationEngine"]




"""
Code Generation Router Implementation
通过代码生成的方式调用环境模块中的@tool标记的接口

架构说明:
- AskContext: 一次ask调用所需的所有状态
- 观察者模式: 流程结束后统一将 context 交给所有 observer 记录
- 管道-过滤器: 整体流程通过 PipelineStage 串联
- 责任链: Code 获取通过 CodeProvider 链实现 (PredefinedCode -> Cache -> LLM)
- CodeStage: 单阶段完成代码获取 + 验证 + 执行
"""

import ast
import asyncio
import inspect
import json
import math
import os
import pickle
import random
import re
import sys
from dataclasses import dataclass, field
from datetime import datetime
from io import StringIO
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Protocol, Tuple

if TYPE_CHECKING:
    from agentsociety2.storage import ReplayWriter

import faiss
import numpy as np
from agentsociety2.config import Config
from agentsociety2.env.base import EnvBase
from agentsociety2.env.benchmark import (
    EnvRouterBenchmarkData,
)
from agentsociety2.env.router_base import RouterBase
from agentsociety2.logger import get_logger
from litellm import AllMessageValues, aembedding

__all__ = ["CodeGenRouter", "AskContext"]


@dataclass
class CacheStats:
    """缓存统计信息"""
    request_count: int = 0  # 总请求次数
    predefined_hit_count: int = 0  # 预定义指令命中次数（observe/statistics）
    cache_hit_count: int = 0  # 缓存命中次数
    cache_miss_count: int = 0  # 缓存未命中次数
    total_input_tokens: int = 0  # 总输入token数
    total_output_tokens: int = 0  # 总输出token数
    code_execution_success_count: int = 0  # 代码执行成功次数
    code_execution_failure_count: int = 0  # 代码执行失败次数
    total_code_retry_count: int = 0  # 总代码重试次数

    @property
    def cache_hit_rate(self) -> float:
        """缓存命中率"""
        total = self.cache_hit_count + self.cache_miss_count
        return self.cache_hit_count / total if total > 0 else 0.0

    @property
    def code_execution_success_rate(self) -> float:
        """代码执行成功率"""
        total = self.code_execution_success_count + self.code_execution_failure_count
        return self.code_execution_success_count / total if total > 0 else 0.0

    @property
    def avg_input_tokens(self) -> float:
        """平均输入token数"""
        return self.total_input_tokens / self.request_count if self.request_count > 0 else 0.0

    @property
    def avg_output_tokens(self) -> float:
        """平均输出token数"""
        return self.total_output_tokens / self.request_count if self.request_count > 0 else 0.0

    @property
    def avg_retry_count(self) -> float:
        """平均代码重试次数"""
        return self.total_code_retry_count / self.request_count if self.request_count > 0 else 0.0


@dataclass
class CacheEntry:
    """缓存条目"""
    instruction_template: str  # 模板指令
    variable_keys: tuple[str, ...]  # 变量键的元组（用于子集检查）
    variable_types: dict[str, str]  # 变量类型字典 {key: type_name}
    code: str  # 生成的代码
    embedding: Optional[np.ndarray] = None  # 指令的embedding（用于相似度计算）
    env_class_type: str = ""  # 接入的env module classes指纹，仅一致时才能使用
    entry_id: Optional[int] = None  # 数据库中的条目ID（持久化时使用）
    success_count: int = 0  # 成功执行次数
    failure_count: int = 0  # 失败执行次数
    last_used: datetime = field(default_factory=datetime.now)  # 最后使用时间
    created_at: datetime = field(default_factory=datetime.now)  # 创建时间

    @property
    def total_usage(self) -> int:
        """总使用次数"""
        return self.success_count + self.failure_count

    @property
    def success_rate(self) -> float:
        """成功率"""
        total = self.success_count + self.failure_count
        return self.success_count / total if total > 0 else 0.0


OBSERVE_INSTRUCTION = "Collect environment observations by calling all available observe tools. For tools that require agent parameters (like agent_id, id, or person_id), extract the agent ID from the ctx dictionary. Store all observation results in results['observations']."
STATISTICS_INSTRUCTION = "Collect environment statistics by calling all available statistics tools. Store all statistics results in results['statistics']."


@dataclass
class AskContext:
    """
    一次ask调用所需的完整状态，在管道各阶段之间传递。
    """
    # === 输入 ===
    ctx: dict
    instruction: str
    readonly: bool
    template_mode: bool

    # === 预处理后 ===
    variables: dict = field(default_factory=dict)
    instruction_stripped: str = ""
    is_observe_or_statistics: bool = False
    resolved_instruction: str = ""  # <observe>/<statistics> 解析后的实际指令

    # === Code 获取 (责任链) ===
    code: Optional[str] = None
    cache_entry: Optional["CacheEntry"] = None
    code_source: Optional[str] = None  # "predefined" | "cache" | "llm"

    # === LLM 重试状态 ===
    retry_count: int = 0
    previous_code: Optional[str] = None
    previous_errors: List[str] = field(default_factory=list)
    dialog_history: List[AllMessageValues] = field(default_factory=list)

    # === 执行结果 ===
    execution_result: Optional[Dict[str, Any]] = None
    execution_attempted: bool = False  # 是否尝试过执行代码（供 observer 统计）
    success_data: Optional[Dict[str, Any]] = None  # 成功时的 {ctx, instruction, results, process_text, status, error, code}

    # === 输出 ===
    final_answer: str = ""
    results: dict = field(default_factory=dict)
    early_return: Optional[Tuple[dict, str]] = None  # 非 None 表示提前返回，不再继续管道
    token_usage_responses: List[Dict[str, int]] = field(default_factory=list)  # 每次 LLM 调用的 token 使用，供 observer 统计

    def __post_init__(self):
        self.instruction_stripped = self.instruction.strip()
        self.is_observe_or_statistics = self.instruction_stripped in ("<observe>", "<statistics>")
        self.variables = self.ctx.get("variables", {})
        self.resolved_instruction = self.instruction
        if self.instruction_stripped == "<observe>":
            self.resolved_instruction = OBSERVE_INSTRUCTION
        elif self.instruction_stripped == "<statistics>":
            self.resolved_instruction = STATISTICS_INSTRUCTION


def _get_debug_info(description: str = "") -> str:
    """获取当前文件、行号和阶段描述的调试信息"""
    frame = inspect.currentframe()
    if frame and frame.f_back:
        caller_frame = frame.f_back
        filename = os.path.basename(caller_frame.f_code.co_filename)
        lineno = caller_frame.f_lineno
        return f"[{filename}:{lineno}] {description}"
    return description


async def clean_coroutines_from_results(results: dict) -> dict:
    """
    递归检查 results 中的未 await 的 coroutine 并 await 获取其值。
    供 CodeStage 和 InstructionLogObserver 使用。
    """
    visited: set[int] = set()

    async def clean_value(value: Any) -> Any:
        if inspect.iscoroutine(value):
            try:
                return await value
            except Exception as e:
                get_logger().warning(f"Failed to await coroutine: {str(e)}")
                return f"<unawaited_coroutine: {type(value).__name__}>"
        elif isinstance(value, dict):
            obj_id = id(value)
            if obj_id in visited:
                return "<circular_reference>"
            visited.add(obj_id)
            try:
                return {k: await clean_value(v) for k, v in value.items()}
            finally:
                visited.remove(obj_id)
        elif isinstance(value, (list, tuple)):
            obj_id = id(value)
            if obj_id in visited:
                return "<circular_reference>"
            visited.add(obj_id)
            try:
                cleaned = [await clean_value(item) for item in value]
                return cleaned if isinstance(value, list) else tuple(cleaned)
            finally:
                visited.remove(obj_id)
        return value

    return await clean_value(results)


def _get_env_class_type_key(env_modules: list) -> str:
    """
    根据 env module classes 生成可读的标识字符串。
    相同 env 配置（模块类型和名称）产生相同 key，用于缓存隔离。
    返回 human-readable 的 JSON，便于直接查看缓存内容。
    """
    keys = []
    for m in env_modules:
        cls = m.__class__
        full_name = f"{cls.__module__}.{cls.__qualname__}"
        keys.append((m.name, full_name))
    return json.dumps(sorted(keys), sort_keys=True, ensure_ascii=False)


class TemplateCacheDB:
    """
    持久化模板缓存，支持跨运行共用。
    使用 pickle 文件存储，FAISS 进行 embedding 相似度搜索。
    缓存按 env_class_type 隔离，仅 env 类型一致时才能命中。
    """

    def __init__(
        self,
        cache_path: str,
        embedding_dims: int,
        max_size_per_env: int = 1000,
    ):
        self._cache_path = cache_path
        self._embedding_dims = embedding_dims
        self._max_size_per_env = max_size_per_env
        os.makedirs(os.path.dirname(cache_path), exist_ok=True)

    def _load_data(self) -> dict:
        """从 pickle 文件加载数据。"""
        if not os.path.exists(self._cache_path):
            return {"next_id": 1, "by_env": {}}
        try:
            with open(self._cache_path, "rb") as f:
                return pickle.load(f)
        except Exception as e:
            get_logger().warning(f"Failed to load cache from {self._cache_path}: {e}")
            return {"next_id": 1, "by_env": {}}

    def _save_data(self, data: dict) -> None:
        """保存数据到 pickle 文件。"""
        with open(self._cache_path, "wb") as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    def load_entries(self, env_class_type: str) -> Tuple[List[CacheEntry], Optional["faiss.Index"], List[int]]:
        """
        加载指定 env_class_type 的缓存条目并构建 FAISS 索引。
        Returns:
            (entries, faiss_index, faiss_entry_indices)
        """
        data = self._load_data()
        raw = data.get("by_env", {}).get(env_class_type, [])
        raw = sorted(raw, key=lambda e: e.last_used, reverse=True)[: self._max_size_per_env]

        entries: List[CacheEntry] = []
        embeddings: List[np.ndarray] = []
        entry_indices: List[int] = []

        for i, e in enumerate(raw):
            e.env_class_type = env_class_type
            entries.append(e)
            if e.embedding is not None:
                embeddings.append(e.embedding)
                entry_indices.append(i)

        if not embeddings:
            return entries, None, []

        emb_array = np.asarray(embeddings, dtype=np.float32).copy()
        faiss.normalize_L2(emb_array)
        index = faiss.IndexFlatIP(len(embeddings[0]))
        index.add(emb_array)
        return entries, index, entry_indices

    def add_entry(
        self,
        env_class_type: str,
        entry: CacheEntry,
    ) -> int:
        """新增缓存条目并持久化。Returns 新条目的 id。"""
        data = self._load_data()
        next_id = data.get("next_id", 1)
        by_env = data.setdefault("by_env", {})
        lst = by_env.setdefault(env_class_type, [])

        entry.entry_id = next_id
        entry.env_class_type = env_class_type
        lst.append(entry)
        lst.sort(key=lambda e: e.last_used, reverse=True)
        by_env[env_class_type] = lst[: self._max_size_per_env]

        data["next_id"] = next_id + 1
        self._save_data(data)
        return next_id

    def update_entry(
        self,
        env_class_type: str,
        entry_id: int,
        success: bool,
        code: str,
    ) -> None:
        """更新已有条目的统计和代码。"""
        data = self._load_data()
        lst = data.get("by_env", {}).get(env_class_type, [])
        for e in lst:
            if e.entry_id == entry_id:
                e.last_used = datetime.now()
                e.code = code
                if success:
                    e.success_count += 1
                else:
                    e.failure_count += 1
                break
        self._save_data(data)

    def clear_env_cache(self, env_class_type: str) -> None:
        """清空指定 env_class_type 的所有缓存条目。"""
        data = self._load_data()
        data.setdefault("by_env", {}).pop(env_class_type, None)
        self._save_data(data)

    def find_by_instruction(
        self,
        env_class_type: str,
        instruction_template: str,
    ) -> Optional[CacheEntry]:
        """按 instruction_template 查找已有条目（用于更新而非新增）。"""
        data = self._load_data()
        lst = data.get("by_env", {}).get(env_class_type, [])
        for e in lst:
            if e.instruction_template == instruction_template:
                return e
        return None


# ==================== 观察者模式 ====================
# 所有 observer 仅在流程结束时被调用一次，根据完整 context 记录必要内容


class AskObserver(Protocol):
    """Ask流程观察者协议：流程结束后接收最终 context"""

    async def on_final(self, context: AskContext) -> None:
        ...


# ==================== 责任链：Code 获取 ====================


class CodeProvider(Protocol):
    """Code 获取责任链节点协议，Predefined 最优先"""

    @property
    def name(self) -> str:
        """提供者名称，用于 context.code_source"""
        ...

    async def get_code(self, context: AskContext, router: "CodeGenRouter") -> Optional[str]:
        """返回代码则链终止，返回 None 则传递至下一节点"""
        ...


# ==================== 管道-过滤器 ====================


class PipelineStage(Protocol):
    """管道阶段协议，接收 AskContext，返回可能被修改的 AskContext"""

    async def process(self, context: AskContext, router: "CodeGenRouter") -> AskContext:
        ...


# --- 观察者具体实现：统一在 on_final 中根据 context 记录 ---


class InstructionLogObserver:
    """记录 instruction 到 instruction_log"""

    def __init__(self, router: "CodeGenRouter"):
        self._router = router

    async def on_final(self, context: AskContext) -> None:
        # 执行阶段已清理 ctx；未执行时（如 early_return）需清理以便 pickle
        ctx = context.ctx if context.execution_attempted else await clean_coroutines_from_results(context.ctx)
        log_entry = EnvRouterBenchmarkData(
            instruction=context.instruction,
            context=ctx,
            readonly=context.readonly,
        )
        async with self._router._instruction_log_lock:
            self._router._instruction_log.append(log_entry)
            try:
                with open(self._router._log_path, "wb") as f:
                    pickle.dump(self._router._instruction_log, f)
            except Exception as e:
                get_logger().warning(f"Failed to pickle instruction log: {str(e)}, skipping file write")


class CacheStatsObserver:
    """根据最终 context 更新 CacheStats 统计"""

    def __init__(self, router: "CodeGenRouter"):
        self._router = router

    async def on_final(self, context: AskContext) -> None:
        async with self._router._cache_stats_lock:
            self._router._cache_stats.request_count += 1
            if context.code_source == "predefined":
                self._router._cache_stats.predefined_hit_count += 1
            elif context.cache_entry:
                self._router._cache_stats.cache_hit_count += 1
            elif not context.is_observe_or_statistics:
                self._router._cache_stats.cache_miss_count += 1
            if context.execution_attempted:
                if context.success_data:
                    self._router._cache_stats.code_execution_success_count += 1
                else:
                    self._router._cache_stats.code_execution_failure_count += 1
                self._router._cache_stats.total_code_retry_count += context.retry_count
            for tu in context.token_usage_responses:
                self._router._cache_stats.total_input_tokens += tu.get("input_tokens", 0)
                self._router._cache_stats.total_output_tokens += tu.get("output_tokens", 0)


class CacheAddObserver:
    """执行成功后根据条件写入缓存"""

    def __init__(self, router: "CodeGenRouter"):
        self._router = router

    @staticmethod
    async def _add_to_cache(router: "CodeGenRouter", instruction: str, variables: dict, code: str, success: bool = True) -> None:
        if not router._template_cache_enabled:
            return
        async with router._template_cache_lock:
            variable_keys = tuple(sorted(variables.keys()))
            variable_types = {k: type(v).__name__ for k, v in variables.items()}
            embedding = await CacheCodeProvider._compute_embedding(router, instruction)
            existing = router._cache_db.find_by_instruction(router._env_class_type_key, instruction)
            if existing and existing.entry_id is not None:
                router._cache_db.update_entry(router._env_class_type_key, existing.entry_id, success, code)
                for e in router._cache_entries:
                    if e.instruction_template == instruction:
                        e.last_used = datetime.now()
                        e.code = code
                        if success:
                            e.success_count += 1
                        else:
                            e.failure_count += 1
                        break
                return
            entry = CacheEntry(
                instruction_template=instruction, variable_keys=variable_keys, variable_types=variable_types,
                code=code, embedding=embedding, env_class_type=router._env_class_type_key,
                success_count=1 if success else 0, failure_count=0 if success else 1,
            )
            new_id = router._cache_db.add_entry(router._env_class_type_key, entry)
            entry.entry_id = new_id
            idx = len(router._cache_entries)
            router._cache_entries.append(entry)
            if embedding is not None:
                router._cache_faiss_entry_indices.append(idx)
                emb = embedding.astype(np.float32).reshape(1, -1).copy()
                faiss.normalize_L2(emb)
                if router._cache_faiss_index is None:
                    router._cache_faiss_index = faiss.IndexFlatIP(emb.shape[1])
                assert router._cache_faiss_index is not None
                router._cache_faiss_index.add(emb)

    async def on_final(self, context: AskContext) -> None:
        if context.success_data is None:
            return
        get_logger().info(f"Try to cache: {context.instruction[:100]}...")
        get_logger().info(f"context.template_mode: {context.template_mode}")
        get_logger().info(f"context.cache_entry: {context.cache_entry}")
        get_logger().info(f"context.is_observe_or_statistics: {context.is_observe_or_statistics}")
        sd = context.success_data
        if context.template_mode and not context.cache_entry and not context.is_observe_or_statistics:
            get_logger().info(f"Adding to cache: {context.instruction[:100]}...")
            await self._add_to_cache(self._router, context.instruction, context.variables, sd["code"], success=True)


# --- Code Provider 责任链具体实现 ---


class PredefinedCodeProvider:
    """预定义代码提供者，优先级最高（<observe> / <statistics>）"""

    @property
    def name(self) -> str:
        return "predefined"

    async def get_code(self, context: AskContext, router: "CodeGenRouter") -> Optional[str]:
        if not context.is_observe_or_statistics:
            return None
        if context.instruction_stripped == "<observe>":
            return router._observe_code if router._observe_code else None
        if context.instruction_stripped == "<statistics>":
            return router._statistics_code if router._statistics_code else None
        return None


class CacheCodeProvider:
    """模板缓存提供者，优先级第二"""

    @property
    def name(self) -> str:
        return "cache"

    @staticmethod
    async def _compute_embedding(router: "CodeGenRouter", text: str) -> Optional[np.ndarray]:
        try:
            async with router._embedding_cache_lock:
                if text in router._embedding_cache:
                    return router._embedding_cache[text]
            response = await aembedding(
                model=f"openai/{router._embedding_model}", input=[text],
                api_key=router._embedding_api_key, api_base=router._embedding_api_base,
            )
            if response and hasattr(response, "data") and len(response.data) > 0:
                emb = np.array(response.data[0]["embedding"], dtype=np.float32)
                async with router._embedding_cache_lock:
                    if len(router._embedding_cache) < 10000:
                        router._embedding_cache[text] = emb
                return emb
            return None
        except Exception as e:
            get_logger().warning(f"Failed to compute embedding: {e}")
            return None

    @staticmethod
    async def _lookup(router: "CodeGenRouter", instruction: str, variables: dict) -> Optional[CacheEntry]:
        if not router._template_cache_enabled:
            return None
        async with router._template_cache_lock:
            emb = await CacheCodeProvider._compute_embedding(router, instruction)
            if emb is None:
                return None
            current_keys = set(variables.keys())
            best_match, best_sim = None, 0.0
            if router._cache_faiss_index and router._cache_faiss_entry_indices:
                query = np.asarray([emb], dtype=np.float32).copy()
                faiss.normalize_L2(query)
                k = min(32, router._cache_faiss_index.ntotal)
                scores, indices = router._cache_faiss_index.search(query, k)
                for score, idx in zip(scores[0], indices[0]):
                    if idx < 0:
                        break
                    entry_idx = router._cache_faiss_entry_indices[idx]
                    entry = router._cache_entries[entry_idx]
                    if entry.env_class_type != router._env_class_type_key:
                        continue
                    cached_keys = set(entry.variable_keys)
                    if not (current_keys.issubset(cached_keys) or cached_keys.issubset(current_keys)):
                        continue
                    sim = float(score)
                    if sim >= router._template_cache_similarity_threshold and sim > best_sim:
                        best_sim, best_match = sim, entry
            if best_match:
                best_match.last_used = datetime.now()
                return best_match
            return None

    async def get_code(self, context: AskContext, router: "CodeGenRouter") -> Optional[str]:
        if not router._template_cache_enabled or not context.template_mode or context.is_observe_or_statistics:
            return None
        cache_entry = await self._lookup(router, context.instruction, context.variables)
        if cache_entry is not None:
            context.cache_entry = cache_entry
            return cache_entry.code
        return None


class LLMCodegenProvider:
    """LLM 代码生成提供者，责任链最后一环，负责调用 LLM 生成代码"""

    @property
    def name(self) -> str:
        return "llm"

    @staticmethod
    def _build_prompt(router: "CodeGenRouter", instruction: str, ctx: dict, readonly: bool, kind: str | None = None) -> str:
        key = (readonly, kind)
        tools_pyi = router._tools_pyi_dict[key]
        return f"""# Code Generation Task
You are a code generation assistant. Your task is to generate Python code that calls environment module tools based on the given agent input.

## Available Environment Modules and Tools
```python
{tools_pyi}
```
```python
modules = {repr(router._modules)}
```

## Agent Input
<instruction>{instruction}</instruction>
```python
ctx = {repr(ctx)}
```

## Code Generation Requirements
1. Generate Python code that accomplishes the instruction by calling appropriate tools.
2. Store results in the `results` dictionary and MUST set results['status'] at the end: 'success', 'in_progress', 'fail', or 'error'.
3. Provide semantic print statements to explain what you're doing.

## Important Notes
- Use `ctx`, `modules`, `results`, `print()`. Allowed modules: collections, itertools, functools, operator, copy, decimal, fractions, statistics, string, re, datetime, json, math, random, numpy (as np).
- Do NOT use dangerous operations. ALWAYS USE `await` TO CALL TOOLS (ASYNC FUNCTIONS).
- NEVER forget to set results['status'] at the END of your code!

## Output Format
Generate ONLY the Python code, without markdown. Start directly with Python statements.

Your generated code:"""

    @staticmethod
    def _build_error_message(previous_errors: List[str]) -> str:
        errors_text = "\n".join(f"- {i+1}. {e}" for i, e in enumerate(previous_errors))
        return f"""The code I generated failed during execution. Here's what went wrong:

## Errors
{errors_text}

Please analyze the errors and fix the code. Common issues: incorrect function/module names, wrong parameter types, async/await usage, type mismatches, missing imports, logic errors.

Please generate the corrected code:"""

    @staticmethod
    async def _call_llm(router: "CodeGenRouter", context: AskContext) -> Tuple[str, Optional[Dict[str, int]]]:
        try:
            response = await router.acompletion_with_system_prompt(model="coder", messages=context.dialog_history)
            raw = response.choices[0].message.content or ""  # type: ignore[union-attr]
            pattern = r"```(?:python)?\s*\n(.*?)```"
            matches = re.findall(pattern, raw, re.DOTALL)
            code = matches[0].strip() if matches else raw.strip()
            usage = getattr(response, "usage", None)
            token_usage = None
            if usage is not None:
                token_usage = {"input_tokens": getattr(usage, "prompt_tokens", 0), "output_tokens": getattr(usage, "completion_tokens", 0)}
            return code, token_usage
        except Exception as e:
            get_logger().error(f"{_get_debug_info('LLM代码生成失败')} - {e}")
            return "", None

    async def get_code(self, context: AskContext, router: "CodeGenRouter") -> Optional[str]:
        if context.retry_count == 0:
            prompt = self._build_prompt(router, context.resolved_instruction, context.ctx, context.readonly, None)
            context.dialog_history = [{"role": "user", "content": prompt}]
        else:
            if context.previous_code:
                context.dialog_history.append({"role": "assistant", "content": context.previous_code})
            context.dialog_history.append({"role": "user", "content": self._build_error_message(context.previous_errors)})
        code, token_usage = await self._call_llm(router, context)
        if token_usage:
            context.token_usage_responses.append(token_usage)
        return code.strip() if code else None


# --- Pipeline 阶段具体实现 ---


class InitStage:
    """初始化：添加时间到 ctx，检查 env_modules"""

    async def process(self, context: AskContext, router: "CodeGenRouter") -> AskContext:
        router._add_current_time_to_ctx(context.ctx)
        if not router.env_modules:
            context.early_return = (context.ctx, "No environment modules available to handle the request.")
        return context


class CodeStage:
    """
    代码获取（责任链）+ 验证 + 执行，含重试循环。
    Predefined -> Cache -> LLM；验证或执行失败时通过 LLM 重试。
    """

    @staticmethod
    async def _acquire_code(router: "CodeGenRouter", context: AskContext, retry_only: bool) -> bool:
        providers = router._code_provider_chain[-1:] if retry_only else router._code_provider_chain
        for provider in providers:
            code = await provider.get_code(context, router)
            if code is not None:
                context.code = code
                context.code_source = provider.name
                return True
        return False

    @staticmethod
    def _validate_code_safety(router: "CodeGenRouter", code: str) -> Tuple[bool, str]:
        violations = []
        try:
            tree = ast.parse(code, mode="exec")
            dangerous_functions = {"eval", "exec", "compile", "__import__", "open", "input"}

            def is_dangerous_module(name: str) -> bool:
                if name in router.ALLOWED_MODULES:
                    return False
                return name in router.DANGEROUS_MODULES or (name.startswith("_") and name != "__future__")

            for node in ast.walk(tree):
                if type(node) in router.FORBIDDEN_AST_NODES:
                    violations.append(f"Forbidden AST node type: {type(node).__name__}")
                elif isinstance(node, ast.Call):
                    if isinstance(node.func, ast.Name) and node.func.id in dangerous_functions:
                        violations.append(f"Dangerous function call: {node.func.id}()")
                    elif isinstance(node.func, ast.Attribute) and node.func.attr in {"eval", "exec", "compile"}:
                        violations.append(f"Dangerous method call: {node.func.attr}()")
                elif isinstance(node, ast.Import):
                    for alias in node.names:
                        if is_dangerous_module(alias.name):
                            violations.append(f"Dangerous import: import {alias.name}")
                elif isinstance(node, ast.ImportFrom) and node.module and is_dangerous_module(node.module):
                    violations.append(f"Dangerous import: from {node.module} import ...")

            if violations:
                return False, "Code safety check failed. Violations found:\n" + "\n".join(f"- {v}" for v in violations)
            return True, ""
        except SyntaxError as e:
            return False, f"Code syntax error: {str(e)}"
        except Exception as e:
            return False, f"Code validation error: {str(e)}"

    @staticmethod
    async def _execute_code(router: "CodeGenRouter", code: str, ctx: dict, readonly: bool) -> Dict[str, Any]:
        results = {}

        def safe_import(name, globals=None, locals=None, fromlist=(), level=0):
            if name not in router.ALLOWED_MODULES:
                return None
            return __import__(name, globals, locals, fromlist, level)

        import collections
        import copy
        import decimal
        import fractions
        import functools
        import itertools
        import operator
        import statistics
        import string
        allowed_modules = {
            "collections": collections, "itertools": itertools, "functools": functools, "operator": operator,
            "copy": copy, "decimal": decimal, "fractions": fractions, "statistics": statistics, "string": string,
            "re": re, "datetime": datetime, "json": json, "math": math, "random": random, "numpy": np, "np": np,
        }
        restricted_builtins = {k: v for k, v in __builtins__.items() if k in router.ALLOWED_BUILTINS}
        restricted_builtins["__import__"] = safe_import
        exec_globals = {
            "__builtins__": restricted_builtins, "ctx": ctx, "modules": router._modules, "results": results,
            "print": print, **allowed_modules,
            "Exception": Exception, "RuntimeError": RuntimeError, "ValueError": ValueError, "TypeError": TypeError,
            "SyntaxError": SyntaxError, "NameError": NameError, "AttributeError": AttributeError,
            "IndexError": IndexError, "KeyError": KeyError,
        }
        exec_locals = {}
        old_stdout = sys.stdout
        sys.stdout = captured_output = StringIO()
        try:
            is_async = "async" in code or "await" in code
            async def run():
                if is_async:
                    indented = "\n".join("    " + line if line.strip() else "" for line in code.split("\n"))
                    async_code = f"async def _generated_main():\n{indented}"
                    exec(compile(async_code, "<generated_async>", "exec"), exec_globals, exec_locals)
                    await exec_locals["_generated_main"]()
                else:
                    exec(compile(code, "<generated>", "exec"), exec_globals, exec_locals)
            await asyncio.wait_for(run(), timeout=10)
            output = captured_output.getvalue()
            print_outputs = [line.strip() for line in output.split("\n") if line.strip()]
            return {"results": results, "output": output, "print_outputs": print_outputs, "success": True}
        except asyncio.TimeoutError:
            raise TimeoutError("Code execution timeout: exceeded 10 seconds limit")
        except Exception as e:
            return {
                "results": results, "output": captured_output.getvalue(), "print_outputs": [],
                "error": str(e), "success": False,
            }
        finally:
            sys.stdout = old_stdout

    async def process(self, context: AskContext, router: "CodeGenRouter") -> AskContext:
        if context.early_return:
            return context
        max_retries = router.max_llm_call_retry if context.code is None else 0
        while context.retry_count <= max_retries:
            if context.code is None:
                ok = await self._acquire_code(router, context, retry_only=bool(context.previous_errors))
                if not ok:
                    context.retry_count += 1
                    context.previous_errors.append("Failed to generate code from LLM.")
                    context.previous_code = None
                    if context.retry_count > max_retries:
                        context.early_return = ({}, "Failed to generate code after retries.")
                        return context
                    continue
                if context.code_source == "llm":
                    context.dialog_history.append({"role": "assistant", "content": context.code})

                assert context.code is not None  # set by _acquire_code when ok=True
                is_safe, safety_violation = self._validate_code_safety(router, context.code)
                if not is_safe:
                    context.retry_count += 1
                    context.previous_code = context.code
                    context.previous_errors.append(safety_violation)
                    if context.retry_count > max_retries:
                        context.early_return = ({}, f"Generated code failed safety check after retries: {safety_violation}")
                        return context
                    context.code = None
                    continue

            code = context.code
            if code is None:
                return context
            context.execution_attempted = True
            try:
                async with router._execute_lock:
                    execution_result = await self._execute_code(router, code, context.ctx, context.readonly)
            except Exception as e:
                execution_result = {"success": False, "error": str(e), "results": {}, "print_outputs": [], "output": ""}

            context.execution_result = execution_result
            # 代码执行后必须清理 ctx 和 results 中的 coroutine，以便序列化
            context.ctx = await clean_coroutines_from_results(context.ctx)
            if execution_result.get("results"):
                execution_result["results"] = await clean_coroutines_from_results(execution_result["results"])
            if not execution_result.get("success", False):
                error = execution_result.get("error", "Unknown error")
                context.retry_count += 1
                context.previous_code = code
                context.previous_errors.append(error)
                if context.retry_count > max_retries:
                    context.early_return = (context.ctx, f"Code execution failed after retries: {error}")
                    return context
                context.code = None
                continue

            print_outputs = execution_result.get("print_outputs", [])
            results = execution_result.get("results", {})
            status = results.get("status", "unknown")
            error = execution_result.get("error", "")
            process_text = "\n".join(print_outputs) if print_outputs else "无输出"
            if error:
                process_text += f"\n\nError: {error}"
            process_text = f"```\n{process_text}\n```"
            context.success_data = {
                "ctx": context.ctx,
                "instruction": context.resolved_instruction,
                "results": results,
                "process_text": process_text,
                "status": status,
                "error": error,
                "code": code,
            }
            context.results = results
            break
        return context


class SummaryStage:
    """生成最终答案"""

    async def process(self, context: AskContext, router: "CodeGenRouter") -> AskContext:
        if context.early_return:
            return context
        if not context.success_data:
            context.early_return = ({}, "Failed to generate and execute code after all retries.")
            return context
        sd = context.success_data
        final_answer, determined_status = await router.generate_final_answer(
            sd["ctx"], sd["instruction"], sd["results"],
            sd["process_text"], sd["status"], sd["error"]
        )
        context.results = sd["results"]
        context.results["status"] = determined_status
        context.final_answer = final_answer
        if determined_status == "unknown":
            context.results["status"] = "fail"
            context.results["reason"] = "Generated code did not set results['status'], which is mandatory"
        return context


class ObserveFinalStage:
    """流程结束后统一将 context 交给所有 observer 记录"""

    async def process(self, context: AskContext, router: "CodeGenRouter") -> AskContext:
        await router._notify_observers_final(context)
        return context


class CodeGenRouter(RouterBase):
    """
    代码生成式Router：通过生成Python代码的方式调用环境模块中的@tool标记的接口。

    工作流程：
    1. 收集所有环境模块的工具信息
    2. 使用类似 pyi 文件的 Python 代码格式向LLM提供环境模块描述和工具信息
       （包含 pydantic BaseModel 定义和模块类定义）
    3. LLM生成Python代码来调用工具
    4. 使用AST解析检查代码安全性
    5. 通过compile和exec执行代码，捕获打印输出
    6. 根据执行结果和打印输出生成最终响应
    """

    # 允许的内置函数
    ALLOWED_BUILTINS = {
        "print",
        "len",
        "str",
        "int",
        "float",
        "bool",
        "list",
        "dict",
        "tuple",
        "set",
        "range",
        "enumerate",
        "zip",
        "min",
        "max",
        "sum",
        "abs",
        "round",
        "sorted",
        "reversed",
        "any",
        "all",
        "isinstance",
        "type",
        "getattr",
        "hasattr",
        "dir",
    }

    # 禁止的AST节点类型（黑名单）
    FORBIDDEN_AST_NODES = {
        ast.ClassDef,  # 禁止定义类
        ast.Delete,  # 禁止del语句
        ast.Global,
        ast.Nonlocal,  # 禁止全局/非局部变量声明
        ast.With,
        ast.AsyncWith,  # 禁止with语句
        ast.Assert,  # 禁止assert
    }

    # 允许导入的模块白名单
    ALLOWED_MODULES = {
        "collections",
        "itertools",
        "functools",
        "operator",
        "copy",
        "decimal",
        "fractions",
        "statistics",
        "string",
        "re",
        "datetime",
        "json",
        "math",
        "random",
        "numpy",
        "np",  # numpy的别名
    }

    # 危险模块列表
    DANGEROUS_MODULES = {
        "os",
        "sys",
        "subprocess",
        "shutil",
        "pickle",
        "marshal",
        "ctypes",
        "socket",
        "urllib",
        "http",
        "ftplib",
        "smtplib",
        "__builtin__",
        "__builtins__",
        "builtins",
    }

    OBSERVE_INSTRUCTION = OBSERVE_INSTRUCTION
    STATISTICS_INSTRUCTION = STATISTICS_INSTRUCTION

    def __init__(
        self,
        env_modules: list[EnvBase],
        max_body_code_lines: int = 10,
        max_steps: int = 10,
        max_llm_call_retry: int = 10,
        log_path: str = "logs/instruction_log.pkl",
        replay_writer: Optional["ReplayWriter"] = None,
        # Template cache configuration
        template_cache_enabled: bool = True,  # 是否启用模板缓存
        template_cache_similarity_threshold: float = 0.85,  # 缓存相似度阈值
        template_cache_max_size: int = 1000,  # 单 env 类型最大缓存条目数
        template_cache_dir: Optional[str] = None,  # 缓存数据库目录，None 则用 {Config.HOME_DIR}/codegen_router_cache
    ):
        super().__init__(
            env_modules=env_modules,
            max_steps=max_steps,
            max_llm_call_retry=max_llm_call_retry,
            replay_writer=replay_writer,
        )

        # Pre-generate all tools pyi code in a dictionary: key is (readonly, kind)
        # kind can be None, "observe", "statistics", etc.
        self._tools_pyi_dict: Dict[Tuple[bool, str | None], str] = {}
        self._log_path = log_path
        os.makedirs(os.path.dirname(log_path), exist_ok=True)

        # Collect all tools info once
        all_tools_info = self._collect_tools_info()

        # Generate writable tools pyi code (kind=None)
        all_tools_info = self._filter_tools_info(
            all_tools_info, readonly=None, kind=None
        )
        self._tools_pyi_dict[(False, None)] = self._format_tools_pyi(
            all_tools_info, max_body_code_lines
        )

        # Generate readonly tools pyi code (kind=None)
        readonly_tools_info = self._filter_tools_info(
            all_tools_info, readonly=True, kind=None
        )
        self._tools_pyi_dict[(True, None)] = self._format_tools_pyi(
            readonly_tools_info, max_body_code_lines
        )

        # Generate readonly observe tools pyi code
        readonly_observe_tools_info = self._filter_tools_info(
            all_tools_info, readonly=True, kind="observe"
        )
        self._tools_pyi_dict[(True, "observe")] = self._format_tools_pyi(
            readonly_observe_tools_info, max_body_code_lines
        )

        # Generate readonly statistics tools pyi code
        readonly_statistics_tools_info = self._filter_tools_info(
            all_tools_info, readonly=True, kind="statistics"
        )
        self._tools_pyi_dict[(True, "statistics")] = self._format_tools_pyi(
            readonly_statistics_tools_info, max_body_code_lines
        )

        self._modules = {module.name: module for module in self.env_modules}

        # Code will be generated using LLM in init method
        self._observe_code = ""
        self._statistics_code = ""

        # Flag to track if LLM code generation has been attempted
        self._llm_code_generated = False

        # 记录所有agent的指令、context和生成的代码
        self._instruction_log: List[EnvRouterBenchmarkData] = []
        self._instruction_log_lock: asyncio.Lock = asyncio.Lock()

        # ==================== Template缓存相关 ====================
        self._template_cache_enabled = template_cache_enabled
        self._template_cache_similarity_threshold = template_cache_similarity_threshold
        self._template_cache_max_size = template_cache_max_size

        # Env class type 指纹，用于缓存隔离
        self._env_class_type_key = _get_env_class_type_key(env_modules)

        # 持久化缓存（pickle 文件，跨运行共用）
        cache_dir = template_cache_dir or os.path.join(Config.HOME_DIR, "codegen_router_cache")
        cache_path = os.path.join(cache_dir, "cache.pkl")
        self._cache_dir = cache_dir  # 与 cache 同目录，用于保存 cache_stats.jsonl
        self._cache_db = TemplateCacheDB(
            cache_path=cache_path,
            embedding_dims=Config.EMBEDDING_DIMS,
            max_size_per_env=template_cache_max_size,
        )
        # 每步统计：用于计算 delta 并写入 JSONL
        self._cache_stats_jsonl_path = os.path.join(cache_dir, "cache_stats.jsonl")
        self._prev_step_stats: Optional[CacheStats] = None
        self._step_index: int = 0
        self._run_id: Optional[str] = None  # 在 init() 时设置，用于区分不同次模拟

        # 当前 env 的缓存集（从 DB 加载，在 init() 时填充）
        self._cache_entries: List[CacheEntry] = []
        self._cache_faiss_index: Optional[faiss.Index] = None
        self._cache_faiss_entry_indices: List[int] = []
        self._template_cache_lock: asyncio.Lock = asyncio.Lock()

        # 缓存统计信息
        self._cache_stats = CacheStats()

        # Embedding缓存
        self._embedding_cache: Dict[str, np.ndarray] = {}

        # Embedding模型配置（从Config获取）
        self._embedding_model = Config.EMBEDDING_MODEL
        self._embedding_api_key = Config.EMBEDDING_API_KEY
        self._embedding_api_base = Config.EMBEDDING_API_BASE
        self._embedding_dims = Config.EMBEDDING_DIMS

        # 并发锁：仅代码执行需串行（会修改 env 状态），其余（缓存、codegen、generate_final_answer）可并行
        self._execute_lock: asyncio.Lock = asyncio.Lock()
        self._cache_stats_lock: asyncio.Lock = asyncio.Lock()
        self._embedding_cache_lock: asyncio.Lock = asyncio.Lock()

        # 观察者（缓存统计、instruction log、缓存写入）
        self._observers: List[AskObserver] = [
            InstructionLogObserver(self),
            CacheStatsObserver(self),
            CacheAddObserver(self),
        ]

        # Code 获取责任链：Predefined -> Cache -> LLM（LLM 在 pipeline 中单独处理）
        self._code_provider_chain: List[CodeProvider] = [
            PredefinedCodeProvider(),
            CacheCodeProvider(),
            LLMCodegenProvider(),
        ]

    async def _notify_observers_final(self, context: AskContext) -> None:
        """流程结束后，将最终 context 交给所有观察者记录"""
        for obs in self._observers:
            await obs.on_final(context)

    async def ask(
        self, ctx: dict, instruction: str, readonly: bool = False, template_mode: bool = False
    ) -> Tuple[dict, str]:
        """
        使用代码生成方式处理指令。通过管道-过滤器架构执行。

        Args:
            ctx: 上下文字典（在template模式下，应包含'variables'键）
            instruction: 指令字符串（在template模式下，为模板指令，包含{variable_name}占位符）
            readonly: 是否只读模式
            template_mode: 是否启用模板模式（启用后使用缓存机制）

        Returns:
            (ctx, answer) 元组
        """
        context = AskContext(ctx=ctx, instruction=instruction, readonly=readonly, template_mode=template_mode)
        stages: List[PipelineStage] = [
            InitStage(),
            CodeStage(),  # 代码获取 + 验证 + 执行
            SummaryStage(),
            ObserveFinalStage(),  # 无论是否 early_return，最后统一交给 observer 记录
        ]
        for stage in stages:
            context = await stage.process(context, self)
        if context.early_return:
            return context.early_return
        return context.results, context.final_answer

    async def init(self, start_datetime: datetime):
        """
        Initialize the router with the start datetime and generate code using LLM.
        从本地缓存数据库加载当前 env 类型的缓存集，构建 FAISS 索引。
        """
        await super().init(start_datetime)

        # 在async上下文中初始化锁
        get_logger().debug("Initialized instruction log lock")

        # 从持久化 DB 加载当前 env_class_type 的缓存，用于 embedding 相似度检索
        if self._template_cache_enabled:
            self._load_cache_from_db()

        # 为本次模拟生成 run_id，便于在 JSONL 中区分不同次模拟
        self._run_id = datetime.now().strftime("%Y%m%d_%H%M%S")

        # Generate code using LLM if not already done
        if not self._llm_code_generated:
            # Generate observe code using LLM (using same logic as regular code generation)
            if (True, "observe") in self._tools_pyi_dict:
                llm_observe_code = await self._generate_observe_code()
                if llm_observe_code:
                    self._observe_code = llm_observe_code
                    get_logger().info("Generated observe code using LLM")
                else:
                    raise ValueError("Failed to generate observe code")
            # Generate statistics code using LLM (using same logic as regular code generation)
            if (True, "statistics") in self._tools_pyi_dict:
                llm_statistics_code = await self._generate_statistics_code()
                if llm_statistics_code:
                    self._statistics_code = llm_statistics_code
                    get_logger().info("Generated statistics code using LLM")
                else:
                    raise ValueError("Failed to generate statistics code")
            self._llm_code_generated = True

    async def step(self, tick: int, t: datetime):
        """
        Run forward one step for all simulation modules, then output cache statistics.
        每步结束后将本步的统计指标以 JSONL 形式追加写入缓存目录，便于分析缓存带来的性能提升。
        """
        await super().step(tick, t)

        # 快照当前累计统计，计算本步增量，写入 JSONL（追加模式，支持多次模拟聚合）
        async with self._cache_stats_lock:
            current = CacheStats(
                request_count=self._cache_stats.request_count,
                predefined_hit_count=self._cache_stats.predefined_hit_count,
                cache_hit_count=self._cache_stats.cache_hit_count,
                cache_miss_count=self._cache_stats.cache_miss_count,
                total_input_tokens=self._cache_stats.total_input_tokens,
                total_output_tokens=self._cache_stats.total_output_tokens,
                code_execution_success_count=self._cache_stats.code_execution_success_count,
                code_execution_failure_count=self._cache_stats.code_execution_failure_count,
                total_code_retry_count=self._cache_stats.total_code_retry_count,
            )
        prev = self._prev_step_stats or CacheStats()
        delta = CacheStats(
            request_count=current.request_count - prev.request_count,
            predefined_hit_count=current.predefined_hit_count - prev.predefined_hit_count,
            cache_hit_count=current.cache_hit_count - prev.cache_hit_count,
            cache_miss_count=current.cache_miss_count - prev.cache_miss_count,
            total_input_tokens=current.total_input_tokens - prev.total_input_tokens,
            total_output_tokens=current.total_output_tokens - prev.total_output_tokens,
            code_execution_success_count=current.code_execution_success_count - prev.code_execution_success_count,
            code_execution_failure_count=current.code_execution_failure_count - prev.code_execution_failure_count,
            total_code_retry_count=current.total_code_retry_count - prev.total_code_retry_count,
        )
        exec_total = delta.code_execution_success_count + delta.code_execution_failure_count
        code_execution_success_rate = (
            delta.code_execution_success_count / exec_total if exec_total > 0 else None
        )
        record = {
            "run_id": self._run_id or "",
            "step": self._step_index,
            "request_count": delta.request_count,
            "predefined_hit_count": delta.predefined_hit_count,
            "cache_hit_count": delta.cache_hit_count,
            "cache_miss_count": delta.cache_miss_count,
            "cached_entries_count": len(self._cache_entries),  # 当前已缓存数据条目标数
            "total_input_tokens": delta.total_input_tokens,
            "total_output_tokens": delta.total_output_tokens,
            "code_execution_success_rate": code_execution_success_rate,
        }
        try:
            os.makedirs(self._cache_dir, exist_ok=True)
            with open(self._cache_stats_jsonl_path, "a", encoding="utf-8") as f:
                f.write(json.dumps(record, ensure_ascii=False) + "\n")
        except Exception as e:
            get_logger().warning(f"Failed to write cache stats JSONL: {e}")

        self._prev_step_stats = current
        self._step_index += 1
        get_logger().info(self.get_cache_stats_summary())

    async def dump(self) -> dict:
        """
        Dump router state to a serializable dict, including instruction logs and cache stats.
        """
        # 调用父类的dump方法获取基础状态
        base_dump = await super().dump()

        # 添加指令日志
        async with self._instruction_log_lock:
            base_dump["instruction_log"] = self._instruction_log.copy()

        # 添加缓存统计信息
        async with self._cache_stats_lock:
            base_dump["cache_stats"] = {
                "request_count": self._cache_stats.request_count,
                "predefined_hit_count": self._cache_stats.predefined_hit_count,
                "cache_hit_count": self._cache_stats.cache_hit_count,
                "cache_miss_count": self._cache_stats.cache_miss_count,
                "total_input_tokens": self._cache_stats.total_input_tokens,
                "total_output_tokens": self._cache_stats.total_output_tokens,
                "code_execution_success_count": self._cache_stats.code_execution_success_count,
                "code_execution_failure_count": self._cache_stats.code_execution_failure_count,
                "total_code_retry_count": self._cache_stats.total_code_retry_count,
            }

        return base_dump

    async def load(self, dump_data: dict):
        """
        Load router state from a dict produced by dump().
        """
        # 调用父类的load方法恢复基础状态
        await super().load(dump_data)

        # 恢复指令日志
        try:
            instruction_log = dump_data.get("instruction_log", [])
            if isinstance(instruction_log, list):
                async with self._instruction_log_lock:
                    self._instruction_log = instruction_log.copy()
                get_logger().debug(
                    f"Loaded {len(instruction_log)} instruction log entries"
                )
        except Exception as e:
            get_logger().warning(f"Failed to load instruction log: {str(e)}")

    async def _generate_initialization_code_with_retry(
        self, instruction: str, ctx: dict, kind: str,
    ) -> str:
        """生成 observe/statistics 初始化代码，使用 LLMCodegenProvider 与 CodeStage。"""
        llm_provider = LLMCodegenProvider()
        code_stage = CodeStage()
        prompt = llm_provider._build_prompt(self, instruction, ctx, True, kind)
        dialog_history: List[AllMessageValues] = [{"role": "user", "content": prompt}]
        previous_code: Optional[str] = None
        previous_errors: List[str] = []
        retry_count = 0

        while retry_count <= self.max_llm_call_retry:
            # 构造临时 context 供 _call_llm 使用
            tmp_ctx = AskContext(ctx=ctx, instruction=instruction, readonly=True, template_mode=False)
            tmp_ctx.dialog_history = dialog_history
            tmp_ctx.retry_count = retry_count
            tmp_ctx.previous_code = previous_code
            tmp_ctx.previous_errors = previous_errors

            code, token_usage = await llm_provider._call_llm(self, tmp_ctx)
            if token_usage:
                async with self._cache_stats_lock:
                    self._cache_stats.total_input_tokens += token_usage.get("input_tokens", 0)
                    self._cache_stats.total_output_tokens += token_usage.get("output_tokens", 0)

            if not code:
                previous_errors.append("Failed to generate code from LLM.")
                previous_code = None
                if retry_count >= self.max_llm_call_retry:
                    raise ValueError(f"Failed to generate {kind} code after retries.")
                retry_count += 1
                if previous_code:
                    dialog_history.append({"role": "assistant", "content": previous_code})
                dialog_history.append({"role": "user", "content": llm_provider._build_error_message(previous_errors)})
                continue

            dialog_history.append({"role": "assistant", "content": code})
            is_safe, safety_violation = code_stage._validate_code_safety(self, code)
            if not is_safe:
                previous_errors.append(safety_violation)
                previous_code = code
                if retry_count >= self.max_llm_call_retry:
                    raise ValueError(f"Generated {kind} code failed safety check after retries: {safety_violation}")
                retry_count += 1
                dialog_history.append({"role": "user", "content": llm_provider._build_error_message(previous_errors)})
                continue

            try:
                async with self._execute_lock:
                    execution_result = await code_stage._execute_code(self, code, ctx, True)
                if not execution_result.get("success", False):
                    previous_errors.append(f"Execution failed: {execution_result.get('error', 'Unknown error')}")
                    previous_code = code
                    if retry_count >= self.max_llm_call_retry:
                        raise ValueError(f"Generated {kind} code failed execution after retries.")
                    retry_count += 1
                    dialog_history.append({"role": "user", "content": llm_provider._build_error_message(previous_errors)})
                    continue
            except Exception as e:
                previous_errors.append(f"Execution exception: {str(e)}")
                previous_code = code
                if retry_count >= self.max_llm_call_retry:
                    raise ValueError(f"Generated {kind} code failed execution after retries: {str(e)}")
                retry_count += 1
                dialog_history.append({"role": "user", "content": llm_provider._build_error_message(previous_errors)})
                continue

            return code.strip()
        raise ValueError(f"Failed to generate {kind} code after retries.")

    async def _generate_observe_code(self) -> str:
        """
        使用LLM生成观察代码，用于调用所有observe类型的工具。
        使用与其他普通文本相同的代码生成逻辑，失败时通过多轮对话让LLM修正。

        Returns:
            生成的Python代码字符串，如果生成失败则返回空字符串
        """
        get_logger().debug(f"{_get_debug_info('开始生成observe代码')}")

        if (True, "observe") not in self._tools_pyi_dict:
            get_logger().debug(f"{_get_debug_info('observe工具不存在')} - 跳过代码生成")
            return ""

        instruction = self.OBSERVE_INSTRUCTION
        ctx = {"id": 123}  # 测试执行用的最小上下文
        return await self._generate_initialization_code_with_retry(
            instruction=instruction, ctx=ctx, kind="observe"
        )

    async def _generate_statistics_code(self) -> str:
        """
        使用LLM生成统计代码，用于调用所有statistics类型的工具。
        使用与其他普通文本相同的代码生成逻辑，失败时通过多轮对话让LLM修正。

        Returns:
            生成的Python代码字符串，如果生成失败则返回空字符串
        """
        get_logger().debug(f"{_get_debug_info('开始生成statistics代码')}")

        if (True, "statistics") not in self._tools_pyi_dict:
            get_logger().debug(
                f"{_get_debug_info('statistics工具不存在')} - 跳过代码生成"
            )
            return ""

        instruction = self.STATISTICS_INSTRUCTION
        ctx = {}  # 测试执行用的最小上下文
        return await self._generate_initialization_code_with_retry(
            instruction=instruction, ctx=ctx, kind="statistics"
        )

    def _load_cache_from_db(self) -> None:
        """从本地缓存数据库加载当前 env_class_type 的缓存集，构建 FAISS 索引。"""
        entries, index, indices = self._cache_db.load_entries(self._env_class_type_key)
        self._cache_entries = entries
        self._cache_faiss_index = index
        self._cache_faiss_entry_indices = indices
        env_preview = self._env_class_type_key[:80] + ("..." if len(self._env_class_type_key) > 80 else "")
        get_logger().info(
            f"Loaded {len(entries)} cache entries (env={env_preview}), FAISS index size: {len(indices)}"
        )

    def get_cache_stats(self) -> CacheStats:
        """
        获取缓存统计信息。

        Returns:
            CacheStats对象，包含所有缓存统计信息
        """
        # 同步方法无法用 async with，暂不加锁；step 中调用 get_cache_stats_summary 为最佳-effort
        return self._cache_stats

    def get_cache_stats_summary(self) -> str:
        """
        获取缓存统计信息的摘要字符串。

        Returns:
            格式化的统计信息字符串
        """
        stats = self._cache_stats
        return f"""Cache Statistics Summary:
- Total Requests: {stats.request_count}
- Predefined Hits (observe/statistics): {stats.predefined_hit_count}
- Cache Hits: {stats.cache_hit_count}
- Cache Misses: {stats.cache_miss_count}
- Cache Hit Rate: {stats.cache_hit_rate:.2%}
- Total Input Tokens: {stats.total_input_tokens}
- Total Output Tokens: {stats.total_output_tokens}
- Average Input Tokens: {stats.avg_input_tokens:.2f}
- Average Output Tokens: {stats.avg_output_tokens:.2f}
- Code Execution Success: {stats.code_execution_success_count}
- Code Execution Failure: {stats.code_execution_failure_count}
- Code Execution Success Rate: {stats.code_execution_success_rate:.2%}
- Total Code Retries: {stats.total_code_retry_count}
- Average Code Retries per Request: {stats.avg_retry_count:.2f}
- Cache Size: {len(self._cache_entries)} entries (env={self._env_class_type_key[:60] + ('...' if len(self._env_class_type_key) > 60 else '')})"""

    async def clear_cache(self) -> None:
        """清空当前 env 的缓存（内存 + 持久化 DB）"""
        async with self._template_cache_lock:
            self._cache_db.clear_env_cache(self._env_class_type_key)
            self._cache_entries = []
            self._cache_faiss_index = None
            self._cache_faiss_entry_indices = []
            self._embedding_cache.clear()
            get_logger().info("Template cache cleared")



"""
Data models for module initialization and configuration.
"""

from datetime import datetime
from typing import Any, Dict, List
from pydantic import BaseModel, Field

__all__ = [
    "EnvModuleInitConfig",
    "AgentInitConfig",
    "CreateInstanceRequest",
    "AskRequest",
    "InterventionRequest",
]


class EnvModuleInitConfig(BaseModel):
    """Configuration for initializing an environment module."""

    module_type: str = Field(
        ...,
        description="The type of environment module (e.g., 'global_information', 'economy_space', 'social_space', 'mobility_space')",
    )

    args: Dict[str, Any] = Field(
        default_factory=dict,
        description="Module initialization arguments (excluding llm). These arguments are passed directly to the module constructor. See mcp_description() for the module's expected parameters and JSON schemas.",
    )


class AgentInitConfig(BaseModel):
    """Configuration for initializing an agent."""

    agent_type: str = Field(
        ..., description="The type of agent (e.g., 'person_agent', 'custom_agent')"
    )

    agent_id: int = Field(..., description="Unique ID for the agent")

    args: Dict[str, Any] = Field(
        default_factory=dict,
        description="Agent initialization arguments (excluding llm and env which are set via init() method). All other parameters including 'id', 'profile', 'memory_config', etc. should be included here. See mcp_description() for the agent's expected parameters.",
    )


class CreateInstanceRequest(BaseModel):
    """Request to create a new AgentSociety instance."""

    instance_id: str = Field(
        ..., description="Unique identifier for this society instance"
    )

    env_modules: List[EnvModuleInitConfig] = Field(
        ..., min_length=1, description="List of environment modules to initialize"
    )

    agents: List[AgentInitConfig] = Field(
        ..., min_length=1, description="List of agents to initialize"
    )

    start_t: datetime = Field(..., description="Simulation start time")

    tick: int = Field(
        1, ge=1, description="Tick duration in seconds for run operations"
    )


class AskRequest(BaseModel):
    """Request to ask or intervene in a society instance."""

    instance_id: str = Field(..., description="The ID of the society instance")

    question: str = Field(..., description="The question to ask")


class InterventionRequest(BaseModel):
    """Request to intervene in a society instance."""

    instance_id: str = Field(..., description="The ID of the society instance")

    instruction: str = Field(..., description="The intervention instruction")
